"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
import argparse
from abc import ABC
from abc import abstractmethod
from programy.utils.console.console import outputLog


class Intent(ABC):

    @abstractmethod
    def generate(self):
        raise NotImplementedError()  # pragma: no cover


class SystemIntent(Intent):

    def __init__(self, samples_file=None):
        self._samples_file = samples_file

    @abstractmethod
    def get_name(self):
        raise NotImplementedError()  # pragma: no cover

    def _load_samples(self):
        samples = []

        if self._samples_file is not None:
            with open(self._samples_file, "r") as samples_file:
                for line in samples_file:
                    samples.append(line.strip())

        return samples

    def generate(self):
        intent = {}
        intent["name"] = self.get_name()
        intent["samples"] = self._load_samples()
        return intent


class CancelIntent(SystemIntent):

    def __init__(self, samples_file=None):
        SystemIntent.__init__(self, samples_file)

    def get_name(self):
        return "AMAZON.CancelIntent"


class HelpIntent(SystemIntent):

    def __init__(self, samples_file=None):
        SystemIntent.__init__(self, samples_file)

    def get_name(self):
        return "AMAZON.HelpIntent"


class StopIntent(SystemIntent):

    def __init__(self, samples_file=None):
        SystemIntent.__init__(self, samples_file)

    def get_name(self):
        return "AMAZON.StopIntent"


class QueryIntent():

    def __init__(self, query):
        self._query = query

    def create_name(self):
        words = self._query.split(" ")
        return "Ask" + "".join([word[0].upper() + word[1:] for word in words])

    def generate(self):
        slot = {}
        slot['name'] = self.create_name()
        slot['slots'] = [{"name": "text", "type": "AMAZON.SearchQuery"}]
        slot['samples'] = ["%s {text}" % self._query]
        return slot


class LanguageModel():

    def __init__(self, invocationName, intents):

        assert isinstance(invocationName, str)
        assert isinstance(intents, list)

        self._invocationName = invocationName
        self._intents = intents

    def generate(self):
        model = {}
        model["invocationName"] = self._invocationName
        model["intents"] = []
        for intent in self._intents:
            model["intents"].append(intent.generate())
        return model


class InteractionModel():

    def __init__(self, languageModel):

        assert isinstance(languageModel, LanguageModel)

        self._languageModel = languageModel

    def generate(self):
        model = {}
        model["languageModel"] = self._languageModel.generate()
        return model


class Intents():

    def __init__(self, interactionModel):
        assert isinstance(interactionModel, InteractionModel)

        self._interactionModel = interactionModel

    def generate(self):
        intents = {}
        intents["interactionModel"] = self._interactionModel.generate()
        return intents


class IntentGenerator():

    def __init__(self, invocation_name, system_intents, intents_word_file, intents_json_file, intents_mapping_file):
        self._invocation_name = invocation_name
        self._system_intents = system_intents
        self._intents_word_file = intents_word_file
        self._intents_json_file = intents_json_file
        self._intents_mapping_file = intents_mapping_file

    def generate(self):
        intents = self._system_intents[:]

        intent_mappings = {}
        with open(self._intents_word_file, "r") as intent_words:
            for line in intent_words:
                word = line.strip().lower()
                intent = QueryIntent(word)
                intents.append(intent)
                intent_mappings[intent.create_name()] = word

        intents = Intents(InteractionModel(LanguageModel(self._invocation_name, intents)))

        with open(self._intents_json_file, "w+") as intent_json:
            intent_json.write(json.dumps(intents.generate(), indent=4, sort_keys=True))

        with open(self._intents_mapping_file, "w+") as intent_json:
            intent_json.write(json.dumps(intent_mappings, indent=4, sort_keys=True))


if __name__ == '__main__':

    outputLog(None, "Generating intents...")

    parser = argparse.ArgumentParser(description='Program-Y Alexa Client Intent Generatorr')
    try:

        parser.add_argument('-in', '--invocation_name', required=True, help="Invocation Name")
        parser.add_argument('-ci', '--cancel_intent', action='store_true', help="Include Cancel Intent")
        parser.add_argument('-cif', '--cancel_intent_file', help="Cancel Intents file")
        parser.add_argument('-hi', '--help_intent', action='store_true', help="Include Help Intent")
        parser.add_argument('-hif', '--help_intent_file', help="Help Intent file")
        parser.add_argument('-si', '--stop_intent', action='store_true', help="Include Stop Intent")
        parser.add_argument('-sif', '--stop_intent_file', help="Stop Intent file")
        parser.add_argument('-if', '--intents_file', required=True, help="Intents filename")
        parser.add_argument('-ij', '--intents_json', required=True, help="Intents JSON filename")
        parser.add_argument('-im', '--intents_maps', required=True, help="Intents Map filename")

        args = parser.parse_args()

        system_intents = []
        if args.cancel_intent is True:
            system_intents.append(CancelIntent(args.cancel_intent_file))
        if args.stop_intent is True:
            system_intents.append(StopIntent(args.stop_intent_file))
        if args.help_intent is True:
            system_intents.append(HelpIntent(args.help_intent_file))

        outputLog(None, "\nReading [%s]" % args.intents_file)

        generator = IntentGenerator(args.invocation_name, system_intents, args.intents_file, args.intents_json,
                                    args.intents_maps)
        generator.generate()

        outputLog(None, "\nGenerated [%s]" % args.intents_json)
        outputLog(None, "Generated [%s]" % args.intents_maps)

    except Exception as excep:
        outputLog(None, "Error generating Alexa intents [%s]" % excep)
        parser.print_help()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import json
import urllib.request
from programy.utils.logging.ylogger import YLogger
from programy.utils.geo.latlong import LatLong


class GeoNamesApi:
    POSTALCODESEARCH = "http://api.geonames.org/postalCodeSearchJSON?postalcode={0}&country={1}&maxRows=10&username={2}"

    def __init__(self):
        self.latlong_response_file = None
        self.account_name = None
        self.country = None

    def check_for_license_keys(self, license_keys):

        if license_keys.has_key('GEO_NAMES_ACCOUNTNAME'):
            self.account_name = license_keys.get_key('GEO_NAMES_ACCOUNTNAME')
        else:
            raise Exception("No valid license key GEO_NAMES_ACCOUNTNAME")

        if license_keys.has_key('GEO_NAMES_COUNTRY'):
            self.country = license_keys.get_key('GEO_NAMES_COUNTRY')
        else:
            raise Exception("No valid license key GEO_NAMES_COUNTRY")

    def _format_postcode(self, postcode):
        return "".join(postcode.split(" "))

    def _format_url(self, postcode):
        postcode = self._format_postcode(postcode)
        return GeoNamesApi.POSTALCODESEARCH.format(postcode, self.country, self.account_name)

    def _get_latlong_for_postcode_response(self, url):

        response = urllib.request.urlopen(url)                          # pragma: no cover
        if response is None:                                            # pragma: no cover
            raise Exception("Invalid url: ", url)                       # pragma: no cover

        content = response.read()                                       # pragma: no cover
        if response is None:                                            # pragma: no cover
            raise Exception("Invalid response from GeoNames")           # pragma: no cover

        return json.loads(content.decode('utf8'))                       # pragma: no cover

    def get_latlong_for_postcode(self, postcode):

        url = self._format_url(postcode)

        data = self._get_latlong_for_postcode_response(url)

        if 'postalCodes' not in data:
            raise Exception("Invalid/Unknown post code")

        if not data['postalCodes']:
            raise Exception("Invalid/Unknown post code")

        if 'lat' not in data['postalCodes'][0] or 'lng' not in data['postalCodes'][0]:
            raise Exception("Invalid/Unknown post code")

        return LatLong(data['postalCodes'][0]['lat'], data['postalCodes'][0]['lng'])



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import wikipedia
from programy.utils.logging.ylogger import YLogger
from programy.services.service import Service


class WikipediaAPI:

    # Provide a summary of a single article
    def summary(self, title, sentences=0, chars=0, auto_suggest=True, redirect=True):
        return wikipedia.summary(title, sentences, chars, auto_suggest, redirect)

    # Provide a list of articles matching the query
    # Use summary to return the neccassary action
    def search(self, query, results=10, suggestion=False):
        return wikipedia.search(query, results, suggestion)


class WikipediaService(Service):

    def __init__(self, config=None, api=None):
        Service.__init__(self, config)

        if api is None:
            self._api = WikipediaAPI()
        else:
            self._api = api

    def ask_question(self, client_context, question: str):
        try:
            words = question.split()
            question = " ".join(words[1:])
            if words[0] == 'SUMMARY':
                search = self._api.summary(question, sentences=1)
            elif words[0] == 'SEARCH':
                results = self._api.search(question)
                search = ", ".join(results)
            else:
                YLogger.error(client_context, "Unknown Wikipedia command [%s]", words[0])
                search = ""
            return search
        except wikipedia.exceptions.DisambiguationError:
            YLogger.error(client_context, "Wikipedia search is ambiguous for question [%s]", question)
        except wikipedia.exceptions.PageError:
            YLogger.error(client_context, "No page on Wikipedia for question [%s]", question)
        except Exception:
            YLogger.error(client_context, "General error querying Wikipedia for question [%s]", question)
        return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from programy.utils.logging.ylogger import YLogger
from programy.storage.factory import StorageFactory


class ProcessorCollection:

    def __init__(self):
        self._processors = []

    @property
    def processors(self):
        return self._processors

    def empty(self):
        self.processors.clear()

    def add_processor(self, processor):
        self._processors.append(processor)

    def process(self, context, word_string):
        for processor in self._processors:
            word_string = processor.process(context, word_string)
        return word_string

    def load(self, storage_factory):
        storage_name = self._get_storage_name()
        if storage_factory.entity_storage_engine_available(storage_name) is True:
            try:
                storage_engine = storage_factory.entity_storage_engine(storage_name)
                processor_store = self._get_store(storage_engine)
                processor_store.load(self)
                return True

            except Exception as e:
                YLogger.exception(self, "Failed to load processors from storage", e)

        return False

    def _get_storage_name(self):
        raise NotImplementedError("Override this in derived class, return StorageFactory.XXX name")  # pragma: no cover

    def _get_store(self, storage_engine):
        raise NotImplementedError("Override this in derived class, return Store from StorageEngine")  # pragma: no cover


class PreProcessorCollection(ProcessorCollection):

    def __init__(self):
        ProcessorCollection.__init__(self)

    def _get_storage_name(self):
        return StorageFactory.PREPROCESSORS

    def _get_store(self, storage_engine):
        return storage_engine.preprocessors_store()


class PostProcessorCollection(ProcessorCollection):

    def __init__(self):
        ProcessorCollection.__init__(self)

    def _get_storage_name(self):
        return StorageFactory.POSTPROCESSORS

    def _get_store(self, storage_engine):
        return storage_engine.postprocessors_store()


class PostQuestionProcessorCollection(ProcessorCollection):

    def __init__(self):
        ProcessorCollection.__init__(self)

    def _get_storage_name(self):
        return StorageFactory.POSTQUESTIONPROCESSORS

    def _get_store(self, storage_engine):
        return storage_engine.postquestionprocessors_store()


##################################################################
#
class Processor(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def process(self, context, word_string):
        raise NotImplementedError()     # pragma: no cover


##################################################################
#
class PreProcessor(Processor):

    def __init__(self):
        Processor.__init__(self)

    @abstractmethod
    def process(self, context, word_string):
        raise NotImplementedError()     # pragma: no cover


##################################################################
#
class PostProcessor(Processor):
    def __init__(self):
        Processor.__init__(self)

    @abstractmethod
    def process(self, context, word_string):
        raise NotImplementedError()     # pragma: no cover


##################################################################
#
class PostQuestionProcessor(Processor):
    def __init__(self):
        Processor.__init__(self)

    @abstractmethod
    def process(self, context, word_string):
        raise NotImplementedError()     # pragma: no cover



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.nodes import NodesStore
from programy.utils.classes.loader import ClassLoader


class FileNodeStore(FileStore, NodesStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        NodesStore.__init__(self)

    def _get_storage_path(self):
        raise NotImplementedError()  # pragma: no cover

    def get_storage(self):
        raise NotImplementedError()  # pragma: no cover

    def _load_file_contents(self, collection, filename):
        YLogger.debug(self, "Loading nodes from file [%s]", filename)
        count = 0
        try:
            with open(filename, "r", encoding="utf-8") as file:
                for line in file:
                    line = line.strip()
                    if line:
                        self.process_config_line(collection, line, filename)
        except FileNotFoundError:
            YLogger.error(self, "File not found [%s]", filename)

        return count

    def process_config_line(self, node_factory, line, filename):
        if self.valid_config_line(line, filename):

            splits = line.split("=")
            node_name = splits[0].strip()
            if node_name in node_factory.nodes:
                YLogger.error(self, "Node already exists in config [%s]", line)
                return False

            class_name = splits[1].strip()
            YLogger.debug(self, "Pre-instantiating %s Node [%s]", node_factory.type, class_name)
            try:
                node_factory.add_node(node_name, ClassLoader.instantiate_class(class_name))
                return True

            except Exception as e:
                YLogger.exception_nostack(self,
                                          "Failed pre-instantiating %s Node [%s]" % (node_factory.type, class_name), e)

        return False

    def valid_config_line(self, line, filename):

        if not line:
            return False

        if line.startswith('#'):
            return False

        if "=" not in line:
            YLogger.error(self, "Config line missing '=' [%s] in [%s]", line, filename)
            return False

        return True


class FilePatternNodeStore(FileNodeStore):

    def __init__(self, storage_engine):
        FileNodeStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.pattern_nodes_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.pattern_nodes_storage


class FileTemplateNodeStore(FileNodeStore):

    def __init__(self, storage_engine):
        FileNodeStore.__init__(self, storage_engine)

    def _get_storage_path(self):
        return self.storage_engine.configuration.template_nodes_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.template_nodes_storage



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.nodes.wildcard import PatternWildCardNode
from programy.parser.pattern.match import Match
from programy.parser.pattern.nodes.base import PatternNode


class PatternOneOrMoreWildCardNode(PatternWildCardNode):
    MATCH_CHARS = ['_', '*']

    def __init__(self, wildcard, userid='*'):
        PatternWildCardNode.__init__(self, wildcard, userid)

    def is_one_or_more(self):
        return True

    @staticmethod
    def is_wild_card(text):
        return bool(text in PatternOneOrMoreWildCardNode.MATCH_CHARS)

    def matching_wildcards(self):
        return PatternOneOrMoreWildCardNode.MATCH_CHARS

    def to_xml(self, client_context, include_user=False):
        string = ""
        if include_user is True:
            string += '<oneormore userid="%s" wildcard="%s">\n' % (self.userid, self.wildcard)
        else:
            string += '<oneormore wildcard="%s">\n' % self.wildcard
        string += super(PatternOneOrMoreWildCardNode, self).to_xml(client_context)
        string += "</oneormore>\n"
        return string

    def to_string(self, verbose=True):
        if verbose is True:
            return "ONEORMORE [%s] [%s] wildcard=[%s]" % (self.userid, self._child_count(verbose), self.wildcard)
        return "ONEORMORE [%s]" % (self.wildcard)

    def equivalent(self, other):
        if other.is_one_or_more():
            if self.userid == other.userid:
                if self._wildcard == other.wildcard:
                    return True
        return False

    def consume(self, client_context, context, words, word_no, match_type, depth, parent=False):
        del parent

        tabs = self.get_tabs(client_context, depth)

        if context.search_time_exceeded() is True:
            YLogger.error(client_context, "%sMax search time [%d]secs exceeded", tabs, context.max_search_timeout)
            return None

        if context.search_depth_exceeded(depth) is True:
            YLogger.error(client_context, "%sMax search depth [%d] exceeded", tabs, context.max_search_depth)
            return None

        if word_no >= words.num_words():
            return None

        word = words.word(word_no)
        YLogger.debug(client_context, "%sWildcard %s matched %s", tabs, self._wildcard, word)
        context_match = Match(match_type, self, word)
        context.add_match(context_match)
        matches_added = 1

        match = self.check_child_is_wildcard(tabs, client_context, context, words, word_no, match_type, depth)
        if match is not None:
            return match

        if self._topic is not None:
            match = self._topic.consume(client_context, context, words, word_no + 1, Match.TOPIC, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched topic, success!", tabs)
                return match

            if words.word(word_no) == PatternNode.TOPIC:
                YLogger.debug(client_context, "%s Looking for a %s, none given, no match found!", tabs,
                              PatternNode.TOPIC)
                return None

        if self._that is not None:
            match = self._that.consume(client_context, context, words, word_no + 1, Match.THAT, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched that, success!", tabs)
                return match

            if words.word(word_no) == PatternNode.THAT:
                YLogger.debug(client_context, "%s Looking for a %s, none given, no match found!", tabs,
                              PatternNode.THAT)
                return None

        word_no += 1
        if word_no >= words.num_words():
            YLogger.debug(client_context, "%sNo more words", tabs)
            return super(PatternOneOrMoreWildCardNode, self).consume(client_context, context, words, word_no,
                                                                     match_type, depth + 1)

        word = words.word(word_no)

        if self._priority_words or self._children:

            ########################################################################
            # Priority nodes
            for child in self._priority_words:

                result = child.equals(client_context, words, word_no)
                if result.matched is True:
                    word_no = result.word_no
                    YLogger.debug(client_context, "%sWildcard child matched %s", tabs, result.matched_phrase)

                    context_match2 = Match(Match.WORD, child, result.matched_phrase)

                    context.add_match(context_match2)
                    matches_added += 1

                    return child.consume(client_context, context, words, word_no + 1, match_type, depth + 1)

            #########################################################################
            # Children nodes
            for child in self._children:

                result = child.equals(client_context, words, word_no)
                if result.matched is True:
                    word_no = result.word_no
                    YLogger.debug(client_context, "%sWildcard child matched %s", tabs, result.matched_phrase)

                    context_match2 = Match(Match.WORD, child, result.matched_phrase)

                    context.add_match(context_match2)
                    matches_added += 1

                    match = child.consume(client_context, context, words, word_no + 1, match_type, depth + 1)
                    if match is not None:
                        return match

            if self.invalid_topic_or_that(tabs, client_context, word, context, matches_added) is True:
                return None

            YLogger.debug(client_context, "%sWildcard %s matched %s", tabs, self._wildcard, word)
            context_match.add_word(word)

            word_no += 1
            word = words.word(word_no)

        YLogger.debug(client_context, "%sNo children, consume words until next break point", tabs)

        while word_no < words.num_words() - 1:
            match = super(PatternOneOrMoreWildCardNode, self).consume(client_context, context, words, word_no,
                                                                      match_type, depth + 1)
            if match is not None:
                return match

            if self.invalid_topic_or_that(tabs, client_context, word, context, matches_added) is True:
                return None

            YLogger.debug(client_context, "%sWildcard %s matched %s", tabs, self._wildcard, word)
            context_match.add_word(word)

            word_no += 1
            word = words.word(word_no)

        YLogger.debug(client_context, "%sWildcard %s matched %s", tabs, self._wildcard, word)
        context_match.add_word(word)

        if word_no == words.num_words() - 1:
            match = super(PatternOneOrMoreWildCardNode, self).consume(client_context, context, words, word_no + 1,
                                                                      match_type, depth + 1)
        else:
            match = super(PatternOneOrMoreWildCardNode, self).consume(client_context, context, words, word_no,
                                                                      match_type, depth + 1)

        if match is not None:
            return match

        context.pop_matches(matches_added)
        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger


class StorageFactory:

    USERS = "users"
    LINKED_ACCOUNTS = "linked_accounts"
    LINKS = "links"

    CATEGORIES = "categories"
    ERRORS = "errors"
    DUPLICATES = "duplicates"
    LEARNF = "learnf"
    CONVERSATIONS = "conversations"

    MAPS = "maps"
    SETS = "sets"
    RDF = "rdf"

    DENORMAL = "denormal"
    NORMAL = "normal"
    GENDER = "gender"
    PERSON = "person"
    PERSON2 = "person2"
    REGEX_TEMPLATES = "regex_templates"

    PROPERTIES = "properties"
    DEFAULTS = "defaults"
    VARIABLES = "variables"

    TWITTER = "twitter"

    SPELLING_CORPUS = "spelling_corpus"
    LICENSE_KEYS = "license_keys"

    TEMPLATE_NODES = "template_nodes"
    PATTERN_NODES = "pattern_nodes"

    BINARIES = "binaries"
    BRAINTREE = "braintree"

    PREPROCESSORS = "preprocessors"
    POSTPROCESSORS = "postprocessors"
    POSTQUESTIONPROCESSORS = "postquestionprocessors"

    USERGROUPS = "usergroups"

    TRIGGERS = "triggers"

    OOBS = "oobs"

    SERVICES = "services"

    def __init__(self):
        self._storage_engines = {}
        self._store_to_engine_map = {}

    def load_engines_from_config(self, configuration):
        for name, config in configuration.storage_configurations.items():
            try:
                engine = config.create_engine()
                if engine is not None:
                    self._storage_engines[name] = engine

            except Exception as e:
                YLogger.exception(None, "Failed to create storage engine [%s]", e, name)

        for store_name, config in configuration.entity_store.items():
            if config in self._storage_engines:
                self._store_to_engine_map[store_name] = self._storage_engines[config]

            else:
                YLogger.error(self, "%s is not a valid storage engine name", config)

    def storage_engine_available(self, name):
        return name in self._storage_engines

    def storage_engine(self, name):
        if self.storage_engine_available(name):
            return self._storage_engines[name]

        else:
            return None

    def entity_storage_engine_available(self, entity_name):
        return entity_name in self._store_to_engine_map

    def entity_storage_engine(self, entity_name):
        if self.entity_storage_engine_available(entity_name):
            return self._store_to_engine_map[entity_name]

        else:
            return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.services.service import Service
from programy.config.brain.service import BrainServiceConfiguration
from programy.services.requestsapi import RequestsAPI


class PannousAPI:

    def __init__(self, request_api=None):
        if request_api is None:
            self._requests_api = RequestsAPI()
        else:
            self._requests_api = request_api

    def ask_question(self, url, question, login):
        payload = {'input': question, 'login': login}

        response = self._requests_api.get(url, params=payload)
        if response is None:
            raise Exception("No response from pannous service")

        json_data = response.json()

        if 'output' not in json_data:
            raise Exception("'output' section missing from pannous json_data")

        if json_data["output"] is None or not json_data["output"]:
            raise Exception("'output' section has no elements in pannous json_data")

        if 'actions' not in json_data["output"][0]:
            raise Exception("'actions' section in output[0] in pannous json_data")

        if 'say' not in json_data["output"][0]['actions']:
            raise Exception("'say' section missing from output[0]['actions'] in pannous json_data")

        if 'text' not in json_data["output"][0]['actions']['say']:
            raise Exception("'text' section missing from output[0]['actions']['say'] in pannous json_data")

        return json_data["output"][0]['actions']['say']['text']


class PannousService(Service):
    LICENSE_KEY = 'PANNOUS_LOGIN'

    def __init__(self, config: BrainServiceConfiguration, api=None):
        Service.__init__(self, config)

        if api is None:
            self.api = PannousAPI()
        else:
            self.api = api

        self.url = None
        if config.url is None:
            raise Exception("Undefined url parameter")
        else:
            self.url = config.url

    def ask_question(self, client_context, question: str):
        try:
            if client_context.client.license_keys.has_key(PannousService.LICENSE_KEY):
                login = client_context.client.license_keys.get_key(PannousService.LICENSE_KEY)
            else:
                YLogger.error(client_context, "No variable %s found in license key file", PannousService.LICENSE_KEY)
                return ""

            return self.api.ask_question(self.url, question, login)

        except Exception as excep:
            YLogger.error(client_context, str(excep))
            return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.parser.exceptions import ParserException
from programy.utils.text.text import TextUtils


class TemplateBotNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._name = None
        self.local = False

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @staticmethod
    def get_bot_variable(client_context, name):
        value = client_context.brain.properties.property(name)
        if value is None:
            YLogger.error(client_context, "No bot property for [%s]", name)

            value = client_context.brain.properties.property("default_property")
            if value is None:
                YLogger.error(client_context, "No value for default_property")

                value = client_context.brain.configuration.defaults.default_get
                if value is None:
                    YLogger.error(client_context, "No value for default default_property, return 'unknown'")
                    value = "unknown"

        return value

    def resolve_to_string(self, client_context):
        name = self.name.resolve(client_context)
        value = TemplateBotNode.get_bot_variable(client_context, name)
        YLogger.debug(client_context, "[%s] resolved to [%s] = [%s]", self.to_string(), name, value)
        return value

    def to_string(self):
        return "[BOT (%s)]" % (self.name.to_string())

    def to_xml(self, client_context):
        xml = "<bot "
        xml += ' name="%s"' % self.name.resolve(client_context)
        xml += " />"
        return xml

    # ######################################################################################################
    # BOT_PROPERTY_EXPRESSION ::==
    # <bot name="PROPERTY"/> |
    # <bot><name>TEMPLATE_EXPRESSION</name></bot>

    def parse_expression(self, graph, expression):
        name_found = False

        if 'name' in expression.attrib:
            self.name = self.parse_attrib_value_as_word_node(graph, expression, 'name')
            name_found = True

        self.parse_text(graph, self.get_text_from_element(expression))

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'name':
                self.name = self.parse_children_as_word_node(graph, child)
                self.local = False
                name_found = True

            else:
                graph.parse_tag_expression(child, self)

            self.parse_text(graph, self.get_tail_from_element(child))

        if name_found is False:
            raise ParserException("Name not found in bot", xml_element=expression)




import sys
import os

class RDFFormatter():

    def __init__(self):
        self._rdf_file = sys.argv[1]

    def get_basename_category(self):
        basename = os.path.basename(self._rdf_file)
        segments = basename.split(".")
        if len(segments) > 1:
            category = segments[0].upper()
        else:
            category = segments.upper()
        return category

    def read_rdf_rdf_file_file(self, category):
        rdfs = {}

        with open(self._rdf_file, "r") as rdf_rdf_file:
            for line in rdf_rdf_file:
                if line:
                    if line.strip().startswith("#") is False:
                        rdf = line.split(":")
                        if len(rdf) > 2:
                            subject = rdf[0].strip().upper()
                            predicate = rdf[1].strip().upper()
                            object = ":".join([x.strip() for x in rdf[2:]])

                            if subject not in rdfs:
                                rdfs[subject] = {}
                                rdfs[subject]['CATEGORY'] = category

                            rdfs[subject][predicate] = object
        return rdfs

    def write_new_rdf_file(self, rdfs):
        temp_filename = self._rdf_file + ".tmp"
        with open(temp_filename, "w+") as rdf_output_tmp:
            for subject in rdfs.keys():
                rdf_output_tmp.write ( "#%s\n"%subject)
                for predicate in rdfs[subject].keys ():
                    rdf_output_tmp.write("%s:%s:%s\n"%(subject, predicate, rdfs[subject][predicate]))
                rdf_output_tmp.write("\n")

        os.remove(self._rdf_file)
        os.rename(temp_filename, self._rdf_file)

    def run(self):

        category = self.get_basename_category()

        rdfs = self.read_rdf_rdf_file_file(category)

        self.write_new_rdf_file(rdfs)


if __name__ == '__main__':

    def run():
        print("Formatting RDF File")
        formatter = RDFFormatter()
        formatter.run()

    run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
from programy.utils.logging.ylogger import YLogger
from programy.extensions.base import Extension


class ClientAdminExtension(Extension):

    def _commands(self):
        return "LIST BOTS, LIST BRAINS, DUMP BRAIN"

    def _list(self, commands, client_context):
        if len(commands) > 1:

            if commands[1] == 'BOTS':
                ids = client_context.client.bot_factory.botids()
                return ", ".join(ids)

            elif commands[1] == 'BRAINS':

                if len(commands) > 2:
                    botid = commands[2]
                    bot = client_context.client.bot_factory.bot(botid)
                    if bot:
                        ids = bot.brain_factory.brainids()
                        return ", ".join(ids)

                    else:
                        return "Invalid Bot Id [%s]" % botid

        return "Invalid LIST command, BOTS or BRAINS only"

    def _dump(self, commands, client_context):
        if commands[1] == 'BRAIN':

            if len(commands) == 4:
                botid = commands[2]
                bot = client_context.client.bot_factory.bot(botid)
                if bot is not None:
                    brainid = commands[3]
                    brain = bot.brain_factory.brain(brainid)
                    if brain is not None:
                        brain.dump_brain_tree(client_context)
                        return "Brain dumped, see config for location"

                    else:
                        return "Invalid Brain Id [%s]" % brainid

                else:
                    return "Invalid Bot Id [%s]" % botid

            else:
                return "Incomplete DUMP BRAIN Command"

        else:
            return "Invalid DUMP command, BRAIN only"

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):
        YLogger.debug(client_context, "Client Admin - [%s]", data)

        try:
            commands = data.split()

            if commands[0] == 'COMMANDS':
                return self._commands()

            elif commands[0] == 'LIST':
                return self._list(commands, client_context)

            elif commands[0] == 'DUMP':
                return self._dump(commands, client_context)

            else:
                return "Invalid Admin Command, LIST or DUMP only"

        except Exception as e:
            YLogger.exception(client_context, "Failed to execute client admin extension", e)

        return "Client Admin Error"



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
import shutil
import re
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.property import PropertyStore


class FilePropertyStore(FileStore, PropertyStore):
    SPLIT_CHAR = ':'
    COMMENT = '#'

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        PropertyStore.__init__(self)

    def _get_storage_path(self):
        return self._storage_engine.configuration.properties_storage.file

    def get_storage(self):
        return self.storage_engine.configuration.properties_storage

    def add_property(self, name, value):
        pass        # pragma: no cover

    def add_properties(self, properties):
        pass        # pragma: no cover

    def get_properties(self):
        property_filepath = self._get_storage_path()
        properties = self.load_properties(property_filepath)
        return properties

    def _process_line(self, line, properties):
        line = line.strip()
        if line.startswith(FilePropertyStore.COMMENT) is False:
            splits = line.split(FilePropertyStore.SPLIT_CHAR)
            if len(splits) > 1:
                key = splits[0].strip()
                val = self.process_value(splits[1:])
                if val is not None:
                    properties[key] = val
                    return True

        return False

    def _load_properties_from_file(self, property_filepath, properties):
        with open(property_filepath, "r") as props_file:
            for line in props_file:
                self._process_line(line, properties)

    def load_properties(self, property_filepath):
        properties = {}
        YLogger.debug(self, "Loading properties from [%s]", property_filepath)
        try:
            self._load_properties_from_file(property_filepath, properties)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to load properties file [%s]", excep, property_filepath)

        return properties

    def process_value(self, splits):
        return (FilePropertyStore.SPLIT_CHAR.join(splits)).strip()

    def _load_file_contents(self, collection, filename):
        properties = self.load_properties(filename)
        for key, value in properties.items():
            collection.add_property(key, value)


class FileRegexStore(FilePropertyStore):

    def __init__(self, storage_engine):
        FilePropertyStore.__init__(self, storage_engine)

    def get_storage(self):
        return self.storage_engine.configuration.regex_storage

    def _get_storage_path(self):
        return self._storage_engine.configuration.regex_storage.file

    def process_value(self, splits):
        try:
            pattern = (FilePropertyStore.SPLIT_CHAR.join(splits)).strip()
            return re.compile(pattern, re.IGNORECASE)
        except Exception:
            YLogger.error(self, "Invalid regex template [%s]", pattern)
            return None

    def get_regular_expressions(self):
        return self.get_properties()


class FileDefaultVariablesStore(FilePropertyStore):

    def __init__(self, storage_engine):
        FilePropertyStore.__init__(self, storage_engine)

    def get_storage(self):
        return self.storage_engine.configuration.defaults_storage

    def _get_storage_path(self):
        return self._storage_engine.configuration.defaults_storage.file

    def get_defaults_values(self):
        return self.get_properties()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import argparse


class ClientArguments():

    def __init__(self, client, parser=None):
        del client
        del parser
        self._bot_root = "."
        self._logging = logging.DEBUG
        self._config_name = "config.yaml"
        self._config_format = "yaml"
        self._no_loop = False
        self._substitutions = None

    def parse_args(self, client):
        pass    # pragma: no cover

    @property
    def bot_root(self):
        return self._bot_root

    @bot_root.setter
    def bot_root(self, root):
        self._bot_root = root

    @property
    def logging(self):
        return self._logging

    @property
    def config_filename(self):
        return self._config_name

    @property
    def config_format(self):
        return self._config_format

    @property
    def noloop(self):
        return self._no_loop

    @property
    def substitutions(self):
        return self._substitutions


class CommandLineClientArguments(ClientArguments):

    def __init__(self, client, parser=None):
        self.args = None
        self._bot_root = None
        self._logging = None
        self._config_name = None
        self._config_format = None
        self._no_loop = False
        self._substitutions = None

        ClientArguments.__init__(self, client)
        if parser is None:
            self.parser = argparse.ArgumentParser()
        else:
            self.parser = parser

        self.parser.add_argument('--bot_root', dest='bot_root', help='root folder for all bot configuration data')
        self.parser.add_argument('--config', dest='config', help='configuration file location')
        self.parser.add_argument('--substitutions', dest='substitutions', help='values to substitute in the config file')
        self.parser.add_argument('--cformat', dest='cformat', help='configuration file format (yaml|json|ini)')
        self.parser.add_argument('--logging', dest='logging', help='logging configuration file')
        self.parser.add_argument('--noloop', dest='noloop', action='store_true', help='do not enter conversation loop')

        client.add_client_arguments(self.parser)

    def parse_args(self, client):
        self.args = self.parser.parse_args()
        self._bot_root = self.args.bot_root
        self._logging = self.args.logging
        self._config_name = self.args.config
        self._config_format = self.args.cformat
        self._no_loop = self.args.noloop
        self._substitutions = self.args.substitutions
        client.parse_args(self, self.args)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from programy.utils.logging.ylogger import YLogger
from programy.utils.console.console import outputLog


class BaseCollection:
    pass    # pragma: no cover


class SingleStringCollection(BaseCollection):

    def __init__(self):
        BaseCollection.__init__(self)
        self._strings = []

    def empty(self):
        self._strings.clear()

    @property
    def strings(self):
        return self._strings

    def load_from_text(self, text):
        lines = text.split("\n")
        count = 0
        for line in lines:
            line = line.strip()
            if len(line) > 0:
                self._strings.append(line)
                count += 1
        return count


class DoubleStringCharSplitCollection(BaseCollection):

    def __init__(self):
        BaseCollection.__init__(self)
        self._pairs = []

    def empty(self):
        self._pairs.clear()

    def remove(self, name=None):
        del name
        self._pairs.clear()

    @property
    def pairs(self):
        return self._pairs

    def add_value(self, key, value):
        key = key.strip()
        value = value.strip()
        self._pairs.append([key, value])

    def set_value(self, key, value):
        key = key.strip()
        value = value.strip()
        for pair in self._pairs:
            if pair[0] == key:
                pair[1] = value

    def has_key(self, key):
        for pair in self._pairs:
            if pair[0] == key:
                return True
        return False

    def value(self, key):
        for pair in self._pairs:
            if pair[0] == key:
                return pair[1]
        return None

    def get_split_char(self):
        return ','

    def load_from_text(self, text):
        lines = text.split("\n")
        count = 0
        for line in lines:
            line = line.strip()
            kvp = line.split(self.get_split_char())
            if len(kvp) > 1:
                key = kvp[0]
                value = self.get_split_char().join(kvp[1:])
                count += 1
                self.add_value(key, value)
        return count


class DoubleStringPatternSplitCollection(BaseCollection):
    RE_OF_SPLIT_PATTERN = re.compile('\"(.*?)\",\"(.*?)\"')

    def __init__(self):
        BaseCollection.__init__(self)
        self._pairs = {}

    @property
    def pairs(self):
        return self._pairs

    def empty(self):
        self._pairs.clear()

    def has_key(self, key):
        for pair in self._pairs.items():
            if pair[0] == key:
                return True
        return False

    def value(self, key):
        if self.has_key(key):
            return self._pairs[key]
        else:
            return None

    def add_to_lookup(self, index, pattern):
        if index in self._pairs:
            YLogger.error(self, "%s = %s already exists in collection", index, pattern)
        self._pairs[index] = pattern

    def replace_by_pattern(self, replacable):
        alreadys = []
        for key, pair in self._pairs.items():
            try:
                pattern = pair[0]
                if pattern.findall(replacable):
                    found = False
                    for already in alreadys:
                        stripped = key.strip()
                        if stripped in already:
                            found = True

                    if found is not True:
                        to_replace = pair[1]
                        to_replace = self.match_case(replacable, to_replace)

                        if pair[1].endswith("."):
                            replacable = pattern.sub(to_replace, replacable)

                        else:
                            replacable = pattern.sub(to_replace+" ", replacable)

                        alreadys.append(pair[1])

            except Exception as excep:
                YLogger.exception(self, "Invalid regular expression [%s]", excep, str(pair[0]))

        return re.sub(' +', ' ', replacable.strip())

    def match_case(self, replacable, to_replace):
        count = 0
        length = len(replacable)

        for i in range(length):
            if replacable[i].isupper():
                count += 1

        if count == length:
            return to_replace.upper()

        if float(count) > float(length)/3.0:
            return to_replace.upper()

        return to_replace.lower()

    def load_from_text(self, text):
        lines = text.split("\n")
        count = 0
        for line in lines:
            line = line.strip()
            split = self.split_line_by_pattern(line, DoubleStringPatternSplitCollection.RE_OF_SPLIT_PATTERN)
            if split is not None:
                key, value = self.process_key_value(split[0], split[1])
                self.add_to_lookup(key, value)
                count += 1
        return count

    @staticmethod
    def split_line_by_pattern(line, pattern):
        line = line.strip()
        if line is not None and line:
            match = pattern.search(line)
            if match is not None:
                lhs = match.group(1)
                rhs = match.group(2)
                return [lhs, rhs]
            YLogger.error(None, "Pattern is bad [%s]", line)
        return None

    @staticmethod
    def normalise_pattern(pattern):
        pattern = pattern.replace("(", r"\(")
        pattern = pattern.replace(")", r"\)")
        pattern = pattern.replace("[", r"\[")
        pattern = pattern.replace("]", r"\]")
        pattern = pattern.replace("{", r"\{")
        pattern = pattern.replace("}", r"\}")
        pattern = pattern.replace("|", r"\|")
        pattern = pattern.replace("^", r"\^")
        pattern = pattern.replace("$", r"\$")
        pattern = pattern.replace("+", r"\+")
        pattern = pattern.replace(".", r"\.")
        pattern = pattern.replace("*", r"\*")
        return pattern

    @staticmethod
    def process_key_value(key, value, userid=None):
        del userid
        pattern_text = DoubleStringPatternSplitCollection.normalise_pattern(key)
        start = pattern_text.lstrip()
        middle = pattern_text
        end = pattern_text.rstrip()
        pattern = "(^%s|%s|%s$)" % (start, middle, end)
        replacement = value.upper()
        return key, [re.compile(pattern, re.IGNORECASE), replacement]



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.nosql.redis.store.redisstore import RedisStore
from programy.storage.entities.conversation import ConversationStore


class RedisConversationStore(RedisStore, ConversationStore):
    CHUNK_SIZE = 5000

    def __init__(self, storage_engine):
        RedisStore.__init__(self, storage_engine)
        ConversationStore.__init__(self)

    def _create_key(self, client_context):
        return "{prefix}:conversation:{clientid}".format(prefix=self._storage_engine.prefix,
                                                         clientid=client_context.client.id)

    def _create_ns_keys(self):
        return "{prefix}:*".format(prefix=self._storage_engine.prefix)

    def empty(self):
        ns_keys = self._create_ns_keys()

        YLogger.info(self, "Emptying Conversation from Redis [%s]", ns_keys)

        cursor = '0'
        while cursor != 0:
            cursor, keys = self._storage_engine.redis.scan(cursor=cursor, match=ns_keys,
                                                           count=RedisConversationStore.CHUNK_SIZE)
            if keys:
                self._storage_engine.redis.delete(*keys)

    def _write_conversation(self, client_context, conversation):
        convo_key = self._create_key(client_context)
        YLogger.debug(self, "Adding conversation [%s]", convo_key)

        json_convo = conversation.to_json()
        json_str = json.dumps(json_convo)

        self.save(convo_key, json_str)

    def store_conversation(self, client_context, conversation, commit=True):
        try:
            self._write_conversation(client_context, conversation)

        except Exception as e:
            YLogger.exception(self, "Failed to save conversation to Redis for clientid [%s]", e,
                              client_context.client.id)

    def _read_conversation(self, client_context, conversation):
        h_key = self._create_key(client_context)
        s_key = self._storage_engine.sessions_set_key

        YLogger.debug(self, "Loading Conversation from cache  %s %s", h_key, s_key)

        # Fetch conversation
        YLogger.debug(self, "Fetching conversation [%s]", h_key)
        json_str = self.get(h_key)
        json_convo = json.loads(json_str)

        conversation.create_from_json(json_convo)

    def load_conversation(self, client_context, conversation):

        try:
            self._read_conversation(client_context, conversation)

        except Exception as e:
            YLogger.exception(self, "Failed to load conversation from Redis for clientid [%s]", e,
                              client_context.client.id)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class ParserException(Exception):
    def __init__(self, message, filename=None, xml_exception=None, xml_element=None):
        Exception.__init__(self, message)
        self._message = message
        self._filename = filename
        self._xml_exception = xml_exception
        self._xml_element = xml_element

    @property
    def message(self):
        return self._message

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename):
        self._filename = filename

    @property
    def xml_exception(self):
        return self._xml_exception

    @xml_exception.setter
    def xml_exception(self, xml_exception):
        self._xml_exception = xml_exception

    @property
    def xml_element(self):
        return self._xml_element

    @xml_element.setter
    def xml_element(self, xml_element):
        self._xml_element = xml_element

    def format_message(self):
        msg = self._message

        if self._filename is not None:
            msg += " in [%s]" % self._filename

        if self._xml_exception is not None:
            if isinstance(self._xml_exception, str):
                msg += " : "
                msg += self._xml_exception
            else:
                msg += " at [line(%d), column(%d)]" % (self._xml_exception.position[0],
                                                       self._xml_exception.position[1])

        if self._xml_element is not None:
            if hasattr(self._xml_element, '_end_line_number') and hasattr(self._xml_element, '_end_column_number'):
                msg += " at [line(%d), column(%d)]" % \
                       (self._xml_element._end_line_number, # pylint: disable=protected-access
                        self._xml_element._end_column_number)  # pylint: disable=protected-access
        return msg


class DuplicateGrammarException(ParserException):
    def __init__(self, message, filename=None, xml_exception=None, xml_element=None):
        ParserException.__init__(self, message, filename=filename, xml_exception=xml_exception, xml_element=xml_element)


class MatcherException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = message



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import socket
import json
from programy.utils.logging.ylogger import YLogger
from programy.clients.events.client import EventBotClient
from programy.clients.events.tcpsocket.config import SocketConfiguration
from programy.clients.render.json import JSONRenderer
from programy.utils.console.console import outputLog


class ClientConnection():

    def __init__(self, clientsocket, addr, max_buffer):
        self._clientsocket = clientsocket
        self._addr = addr
        self._max_buffer = max_buffer

    def receive_data(self):
        json_data = self._clientsocket.recv(self._max_buffer).decode()
        YLogger.debug(self, "Received: %s", json_data)
        return json.loads(json_data, encoding="utf-8")

    def send_response(self, userid, answer):
        return_payload = {"result": "OK", "answer": answer, "userid": userid}
        json_data = json.dumps(return_payload)

        YLogger.debug(self, "Sent %s:", json_data)

        self._clientsocket.send(json_data.encode('utf-8'))

    def send_error(self, error):
        if hasattr(error, 'message') is True:
            return_payload = {"result": "ERROR", "message": error.message}
        elif hasattr(error, 'msg') is True:
            return_payload = {"result": "ERROR", "message": error.msg}
        else:
            return_payload = {"result": "ERROR", "message": str(error)}

        json_data = json.dumps(return_payload)

        YLogger.debug(self, "Sent: %s", json_data)

        self._clientsocket.send(json_data.encode('utf-8'))

    def close(self):
        if self._clientsocket is not None:
            self._clientsocket.close()
            self._clientsocket = None


class SocketFactory():

    def create_socket(self, family=socket.AF_INET, sockettype=socket.SOCK_STREAM):
        return socket.socket(family, sockettype)


class SocketConnection():

    def __init__(self, host, port, queue, max_buffer=1024, factory=SocketFactory()):
        self._socket_connection = self._create_socket(host, port, queue, factory)
        self._max_buffer = max_buffer

    def _create_socket(self, host, port, queue, factory):
        # create a socket object
        serversocket = factory.create_socket()
        # bind to the port
        serversocket.bind((host, port))
        # queue up to 5 requests
        serversocket.listen(queue)
        return serversocket

    def accept_connection(self):
        # establish a connection
        clientsocket, addr = self._socket_connection.accept()
        return ClientConnection(clientsocket, addr, self._max_buffer)


class SocketBotClient(EventBotClient):

    def __init__(self, argument_parser=None):
        self._host = None
        self._port = None
        self._queue = None
        self._max_buffer = None
        self._server_socket = None
        self._client_connection = None

        EventBotClient.__init__(self, "Socket", argument_parser)

        if self._host is not None and \
                self._port is not None and \
                self._queue is not None:
            outputLog(self, "TCP Socket Client server now listening on %s:%d" % (self._host, self._port))
            self._server_socket = self.create_socket_connection(self._host, self._port, self._queue, self._max_buffer)
            self._renderer = JSONRenderer(self)

        else:
            outputLog(self, "TCP Socket Client configuration not defined, unable to create socket!")

    def get_client_configuration(self):
        return SocketConfiguration()

    def parse_configuration(self):
        self._host = self.configuration.client_configuration.host
        self._port = self.configuration.client_configuration.port
        self._queue = self.configuration.client_configuration.queue
        self._max_buffer = self.configuration.client_configuration.max_buffer

    def extract_question(self, receive_payload):
        question = None
        if 'question' in receive_payload:
            question = receive_payload['question']
        if question is None or question == "":
            raise Exception("question missing from payload")
        return question

    def extract_userid(self, receive_payload):
        userid = None
        if 'userid' in receive_payload:
            userid = receive_payload['userid']
        if userid is None or userid == "":
            raise Exception("userid missing from payload")
        return userid

    def create_socket_connection(self, host, port, queue, max_buffer):
        return SocketConnection(host, port, queue, max_buffer)

    def process_question(self, client_context, question):
        self._questions += 1
        response = client_context.bot.ask_question(client_context, question, responselogger=self)
        return self.renderer.render(client_context, response)

    def render_response(self, client_context, response):
        # Calls the renderer which handles RCS context, and then calls back to the client to show response
        self._renderer.render(client_context, response)

    def process_response(self, client_context, response):
        if self._client_connection is not None:
            self._client_connection.send_response(client_context.userid, response)

        else:
            YLogger.error(client_context, "Not client connection, unable to process response")

    def wait_and_answer(self):
        running = True
        try:
            if self._server_socket is None:
                raise Exception("No server socket, unable to accept connections")

            self._client_connection = self._server_socket.accept_connection()

            receive_payload = self._client_connection.receive_data()

            question = self.extract_question(receive_payload)
            userid = self.extract_userid(receive_payload)

            client_context = self.create_client_context(userid)
            answer = self.process_question(client_context, question)

            self.render_response(client_context, answer)

        except KeyboardInterrupt:
            running = False
            YLogger.debug(self, "Cleaning up and exiting...")

        except Exception as e:
            if self._client_connection is not None:
                self._client_connection.send_error(e)

        finally:
            if self._client_connection is not None:
                self._client_connection.close()

        return running


if __name__ == '__main__':
    outputLog(None, "Initiating TCP Socket Client...")


    def run():
        tcpsocket_app = SocketBotClient()
        tcpsocket_app.run()


    run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from flask import Flask, request
from twilio.rest import Client
from twilio.twiml.messaging_response import MessagingResponse
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.twilio.config import TwilioConfiguration
from programy.utils.console.console import outputLog


TWILIO_CLIENT = None


class TwilioBotClient(FlaskRestBotClient):
    
    def __init__(self, argument_parser=None):
        FlaskRestBotClient.__init__(self, 'twilio', argument_parser)

        YLogger.debug(self, "Twilio Client is running....")

        self.get_license_keys()

        self._twilio_client = self.create_twilio_client()

        outputLog(self, "Twilio Client loaded")

    @property
    def from_number(self):
        return self._from_number

    def get_license_keys(self):
        self._account_sid = self.license_keys.get_key("TWILIO_ACCOUNT_SID")
        self._auth_token = self.license_keys.get_key("TWILIO_AUTH_TOKEN")
        self._from_number = self.license_keys.get_key("TWILIO_FROM_NUMBER")

    def get_client_configuration(self):
        return TwilioConfiguration()

    def create_twilio_client(self):
        return Client(self._account_sid, self._auth_token)

    def create_twilio_response(self, client_number, answer):
        response = MessagingResponse()
        response.message(body=answer, to=client_number)
        response_str = str(response)
        return response_str

    def receive_message(self, request):

        if self.configuration.client_configuration.debug is True:
            self.dump_request(request)

        if request.method == 'POST':
            client_number = request.form['From']
            question = request.form['Body']
        else:
            client_number = request.args.get('From')
            question = request.args.get('Body')

        YLogger.debug(self, "Twillio received [%s] from [%s]", question, client_number)

        client_context = self.create_client_context(client_number)

        answer = self.ask_question(client_number, question)
        rendered = self.renderer.render(client_context, answer)

        response = self.create_twilio_response(client_number, rendered)

        YLogger.debug(self, "Twillio sending [%s] to [%s]", answer, client_number)

        return response


if __name__ == "__main__":

    outputLog(None, "Initiating Twilio Client...")

    TWILIO_CLIENT = TwilioBotClient()

    APP = Flask(__name__)

    outputLog(None, "Exposing endpoint: "+TWILIO_CLIENT.configuration.client_configuration.api)
    @APP.route(TWILIO_CLIENT.configuration.client_configuration.api, methods=['POST'])
    def receive_message():
        try:
            return TWILIO_CLIENT.receive_message(request)

        except Exception as excep:
            YLogger.exception(None, "Twilio Error", excep)

    TWILIO_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.entities.store import Store
from programy.storage.stores.nosql.mongo.store.mongostore import MongoStore
from programy.storage.entities.spelling import SpellingStore
from programy.storage.stores.nosql.mongo.dao.corpus import Corpus
from programy.utils.console.console import outputLog


class MongoSpellingStore(MongoStore, SpellingStore):
    SPELLING = 'spelling'

    def __init__(self, storage_engine):
        MongoStore.__init__(self, storage_engine)
        SpellingStore.__init__(self)

    def collection_name(self):
        return MongoSpellingStore.SPELLING

    def _read_corpus_from_file(self, filename, verbose):
        corpus_words = []
        with open(filename, "r") as text_file:
            for lines in text_file:
                words = lines.split(' ')
                for word in words:
                    corpus_words.append(word)

        corpus = Corpus(words=corpus_words)
        if self.add_document(corpus) is False:
            return 0, 0

        count = len(corpus_words)
        return count, count

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        YLogger.info(self, "Uplading spelling corpus file [%s] to Mongo", filename)

        count = success = 0
        try:
            count, success = self._read_corpus_from_file(filename, verbose)

        except Exception as excep:
            YLogger.exception(self, "Failed to load spelling corpus from [%s]", excep, filename)

        # Assume all words loaded are success, no need for additional count
        return count, success

    def load_spelling(self, spell_checker):
        YLogger.info(self, "Loading spelling corpus from Mongo")

        collection = self.collection()
        corpus = collection.find_one({})
        if corpus is not None:
            spell_checker.add_corpus(" ".join(corpus['words']))



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
from programy.clients.events.console.client import ConsoleBotClient
from programy.config.programy import ProgramyConfiguration
from programy.clients.args import CommandLineClientArguments
from programy.utils.substitutions.substitues import Substitutions
from programy.storage.config import FileStorageConfiguration
from programy.storage.stores.file.store.config import FileStoreConfiguration
from programy.storage.factory import StorageFactory


class EmbeddedDataFileBot(ConsoleBotClient):

    def __init__(self, files={}, defaults=False, logging_filename=None):
        self._files = files
        self._defaults = defaults

        filepath = os.path.dirname(__file__) + os.sep
        if logging_filename is not None:
            self._logging_filename = logging_filename
        else:
            self._logging_filename = filepath + 'basicbot/logging.yaml'

        ConsoleBotClient.__init__(self, "Console")

    def _render_callback(self):
        return False

    def parse_arguments(self, argument_parser):
        client_args = CommandLineClientArguments(self, parser=None)
        if self._logging_filename is not None:
            client_args._logging = self._logging_filename
        return client_args

    #def initiate_logging(self, arguments):
    #    pass

    def load_configuration(self, arguments, subs: Substitutions = None):
        client_config = self.get_client_configuration()

        client_config.storage.storage_configurations['file'] = FileStorageConfiguration()

        filepath = os.path.dirname(__file__) + os.sep

        client_config.storage.entity_store[StorageFactory.PATTERN_NODES] = 'file'
        if 'patterns' in self._files:
            client_config.storage.storage_configurations['file']._pattern_nodes_storage = FileStoreConfiguration(file=self._files['patterns'], fileformat="text")
        else:
            # Default if pattern node file not specified
            client_config.storage.storage_configurations['file']._pattern_nodes_storage = FileStoreConfiguration(file=filepath + 'basicbot/nodes/pattern_nodes.conf', fileformat="text")

        client_config.storage.entity_store[StorageFactory.TEMPLATE_NODES] = 'file'
        if 'templates' in self._files:
            client_config.storage.storage_configurations['file']._template_nodes_storage = FileStoreConfiguration(file=self._files['templates'], fileformat="text")
        else:
            # Default if template node file not specified
            client_config.storage.storage_configurations['file']._template_nodes_storage = FileStoreConfiguration(file=filepath + 'basicbot/nodes/template_nodes.conf', fileformat="text")

        if 'aiml' in self._files:
            client_config.storage.entity_store[StorageFactory.CATEGORIES] = 'file'
            client_config.storage.storage_configurations['file']._categories_storage = FileStoreConfiguration(dirs=self._files['aiml'], fileformat="xml",
                                                                                                              extension="aiml", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)

        client_config.storage.entity_store[StorageFactory.LEARNF] = 'file'
        if 'learnf' in self._files:
            client_config.storage.storage_configurations['file']._learnf_storage = FileStoreConfiguration(dirs=self._files['learnf'], fileformat="xml",
                                                                                                              extension="aiml", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._learnf_storage = FileStoreConfiguration(dirs=filepath + 'basicbot/categories/learnf', fileformat="xml",
                                                                                                              extension="aiml", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)

        client_config.storage.entity_store[StorageFactory.PROPERTIES] = 'file'
        if 'properties' in self._files:
            client_config.storage.storage_configurations['file']._properties_storage = FileStoreConfiguration(file=self._files['properties'], fileformat="text")
            
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._properties_storage = FileStoreConfiguration(
                file=filepath + 'basicbot/properties/properties.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.DEFAULTS] = 'file'
        if 'defaults' in self._files:
            client_config.storage.storage_configurations['file']._defaults_storage = FileStoreConfiguration(file=self._files['defaults'], fileformat="text")
        
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._defaults_storage = FileStoreConfiguration(
                file=filepath + 'basicbot/properties/defaults.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.SETS] = 'file'
        if 'sets' in self._files:
            client_config.storage.storage_configurations['file']._sets_storage = FileStoreConfiguration(dirs=self._files['sets'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._sets_storage = FileStoreConfiguration(dirs=[filepath + 'basicbot/sets'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)

        client_config.storage.entity_store[StorageFactory.MAPS] = 'file'
        if 'maps' in self._files:
            client_config.storage.storage_configurations['file']._maps_storage = FileStoreConfiguration(dirs=self._files['maps'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._maps_storage = FileStoreConfiguration(dirs=['basicbot/maps'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)

        client_config.storage.entity_store[StorageFactory.RDF] = 'file'
        if 'rdfs' in self._files:
            client_config.storage.storage_configurations['file']._rdfs_storage = FileStoreConfiguration(dirs=self._files['rdfs'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._rdfs_storage = FileStoreConfiguration(dirs=['basicbot/rdfs'], fileformat="text",
                                                                                                              extension="txt", encoding="utf-8",
                                                                                                              subdirs=True,
                                                                                                              delete_on_start=False)

        client_config.storage.entity_store[StorageFactory.DENORMAL] = 'file'
        if 'denormals' in self._files:
            client_config.storage.storage_configurations['file']._denormal_storage = FileStoreConfiguration(file=self._files['denormals'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._denormal_storage = FileStoreConfiguration(file='basibot/lookups/denormals.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.NORMAL] = 'file'
        if 'normals' in self._files:
            client_config.storage.storage_configurations['file']._normal_storage = FileStoreConfiguration(file=self._files['normals'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._normal_storage = FileStoreConfiguration(file='basicbot/lookups/normals.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.GENDER] = 'file'
        if 'genders' in self._files:
            client_config.storage.storage_configurations['file']._gender_storage = FileStoreConfiguration(file=self._files['genders'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._gender_storage = FileStoreConfiguration(
                file='basicbot/lookups/genders.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.PERSON] = 'file'
        if 'persons' in self._files:
            client_config.storage.storage_configurations['file']._person_storage = FileStoreConfiguration(file=self._files['persons'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._person_storage = FileStoreConfiguration(
                file='basicbot/lookups/persons.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.PERSON2] = 'file'
        if 'person2s' in self._files:
            client_config.storage.storage_configurations['file']._person2_storage = FileStoreConfiguration(file=self._files['person2s'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._person2_storage = FileStoreConfiguration(file='basicbot/lookups/person2s.txt', fileformat="text")

        client_config.storage.entity_store[StorageFactory.REGEX_TEMPLATES] = 'file'
        if 'regexes' in self._files:
            client_config.storage.storage_configurations['file']._regex_storage = FileStoreConfiguration(file=self._files['regexes'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._regex_storage = FileStoreConfiguration(
                file='basicbot/regex/regex-templates.txt', fileformat="text")

        if 'triggers' in self._files:
            client_config.storage.entity_store[StorageFactory.TRIGGERS] = 'file'
            client_config.storage.storage_configurations['file']._triggers_storage = FileStoreConfiguration(file=self._files['triggers'], fileformat="text")

        if 'usergroups' in self._files:
            client_config.storage.entity_store[StorageFactory.USERGROUPS] = 'file'
            client_config.storage.storage_configurations['file']._usergroups_storage = FileStoreConfiguration(file=self._files['usergroups'], fileformat="text")

        if 'spellings' in self._files:
            client_config.storage.entity_store[StorageFactory.SPELLING_CORPUS] = 'file'
            client_config.storage.storage_configurations['file']._spelling_storage = FileStoreConfiguration(file=self._files['spellings'], fileformat="text")

        client_config.storage.entity_store[StorageFactory.PREPROCESSORS] = 'file'
        if 'preprocessors' in self._files:
            client_config.storage.storage_configurations['file']._preprocessors_storage = FileStoreConfiguration(file=self._files['preprocessors'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._preprocessors_storage = FileStoreConfiguration(file=filepath + 'basicbot/processing/preprocessors.conf', fileformat="text")

        client_config.storage.entity_store[StorageFactory.POSTPROCESSORS] = 'file'
        if 'postprocessors' in self._files:
            client_config.storage.storage_configurations['file']._postprocessors_storage = FileStoreConfiguration(file=self._files['postprocessors'], fileformat="text")
        elif self._defaults is True:
            client_config.storage.storage_configurations['file']._postprocessors_storage = FileStoreConfiguration(file=filepath + 'basicbot/processing/postprocessors.conf', fileformat="text")

        if 'postquestionprocessors' in self._files:
            client_config.storage.entity_store[StorageFactory.POSTQUESTIONPROCESSORS] = 'file'
            client_config.storage.storage_configurations['file']._postquestionprocessors_storage = FileStoreConfiguration(file=self._files['postquestionprocessors'], fileformat="text")

        if 'licenses' in self._files:
            client_config.storage.entity_store[StorageFactory.LICENSE_KEYS] = 'file'
            client_config.storage.storage_configurations['file']._license_storage = FileStoreConfiguration(file=self._files['licenses'], fileformat="text")

        if 'errors' in self._files:
            client_config.storage.entity_store[StorageFactory.ERRORS] = 'file'
            client_config.storage.storage_configurations['file']._errors_storage = FileStoreConfiguration(file=self._files['errors'], fileformat="text")

        if 'duplicates' in self._files:
            client_config.storage.entity_store[StorageFactory.ERRORS] = 'file'
            client_config.storage.storage_configurations['file']._duplicates_storage = FileStoreConfiguration(file=self._files['duplicates'], fileformat="text")

        if 'binaries' in self._files:
            client_config.storage.entity_store[StorageFactory.BINARIES] = 'file'
            client_config.storage.storage_configurations['file']._binaries_storage = FileStoreConfiguration(file=self._files['binaries'], fileformat="text")

        if 'braintrees' in self._files:
            client_config.storage.entity_store[StorageFactory.BRAINTREE] = 'file'
            client_config.storage.storage_configurations['file']._braintree_storage = FileStoreConfiguration(file=self._files['braintrees'], fileformat="text")

        if 'conversations' in self._files:
            client_config.storage.entity_store[StorageFactory.CONVERSATIONS] = 'file'
            client_config.storage.storage_configurations['file']._conversation_storage = FileStoreConfiguration(dirs=self._files['conversations'], fileformat="text",
                                                                                                                extension="txt", encoding="utf-8",
                                                                                                                subdirs=True,
                                                                                                                delete_on_start=False)

        if 'oobs' in self._files:
            client_config.storage.entity_store[StorageFactory.OOBS] = 'file'
            client_config.storage.storage_configurations['file']._oobs_storage = FileStoreConfiguration(file=self._files['oobs'], fileformat="text")

        self._configuration = ProgramyConfiguration(client_config)

    def ask_question(self, question):
        client_context = self.create_client_context("testuser")
        return self.renderer.render(client_context, self.process_question(client_context, question))


if __name__ == '__main__':

    files1 = {'aiml'         : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/categories']}

    files2 = {'aiml'         : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/categories'],
             'learnf'       : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/categories/learnf'],
             'properties'   : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/properties/properties.txt',
             'defaults'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/properties/defaults.txt',
             'sets'         : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/sets'],
             'maps'         : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/maps'],
             'rdfs'         : ['/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/rdfs'],
             'denormals'    : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/lookups/denormal.txt',
             'normals'      : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/lookups/normal.txt',
             'genders'      : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/lookups/gender.txt',
             'persons'      : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/lookups/person.txt',
             'person2s'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/lookups/person2.txt',
             'triggers'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/triggers/triggers.txt',
             'regexes'      : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/regex/regex-templates.txt',
             'usergroups'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/security/usergroups.yaml',
             'spellings'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/spelling/corpus.txt',
             'preprocessors': '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/processing/preprocessors.conf',
             'postprocessors': '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/processing/postprocessors.conf',
             'postquestionprocessors': '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/processing/postquestionprocessors.conf',
             'licenses'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/licenses/license.keys',
             'conversations': '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/conversations',
             'duplicates'   : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/debug/duplicates.txt',
             'errors'       : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/debug/errors.txt',
             'services'     : '/Users/keith/Documents/Development/Python/Projects/AIML/y-bot/storage/services',
             }

    print("Loading Bot Brain....please wait!")
    my_bot = EmbeddedDataFileBot(files1, defaults=True)

    print("Asked 'Hello', Response '%s'" % my_bot.ask_question("Hello"))
    print("Asked 'What are you', Response '%s'" % my_bot.ask_question("What are you"))
    print("Asked 'Where are you',  Response '%s'" % my_bot.ask_question("Where are you"))




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.parser.exceptions import ParserException
from programy.parser.pattern.match import Match
from programy.parser.pattern.nodes.base import PatternNode


class PatternRootNode(PatternNode):

    def __init__(self, userid='*'):
        PatternNode.__init__(self, userid)

    def is_root(self):
        return True

    def can_add(self, new_node: PatternNode):
        if new_node.is_root():
            raise ParserException("Cannot add root node to existing root node")
        if new_node.is_topic():
            raise ParserException("Cannot add topic node to root node")
        if new_node.is_that():
            raise ParserException("Cannot add that node to root node")
        if new_node.is_template():
            raise ParserException("Cannot add template node to root node")

    def equivalent(self, other: PatternNode) -> bool:
        if other.is_root():
            if self.userid == other.userid:
                return True
        return False

    def to_string(self, verbose: bool = True) -> str:
        if verbose is True:
            return "ROOT [%s] [%s]" % (self.userid, self._child_count(verbose))
        return "ROOT "

    def match(self, client_context, context, words):
        return self.consume(client_context, context, words, 0, Match.WORD, 0)

    def _remove_priority(self, userid):
        removals = []
        for node in self._priority_words:
            if node.userid == userid:
                removals.append(node)
        for node in removals:
            self._remove_node(node)

    def _remove_0ormore_hash(self, userid):
        if self._0ormore_hash is not None:
            if self._0ormore_hash.userid == userid:
                self._remove_node(self._0ormore_hash)

    def _remove_1ormore_underline(self, userid):
        if self._1ormore_underline is not None:
            if self._1ormore_underline.userid == userid:
                self._remove_node(self._1ormore_underline)

    def _remove_children(self, userid):
        removals = []
        for node in self._children:
            if node.userid == userid:
                removals.append(node)
        for node in removals:
            self._remove_node(node)

    def _remove_0ormore_arrow(self, userid):
        if self._0ormore_arrow is not None:
            if self._0ormore_arrow.userid == userid:
                self._remove_node(self._0ormore_arrow)

    def _remove_1ormore_star(self, userid):
        if self._1ormore_star is not None:
            if self._1ormore_star.userid == userid:
                self._remove_node(self._1ormore_star)

    def remove_children_with_userid(self, userid):

        self._remove_priority(userid)

        self._remove_0ormore_hash(userid)

        self._remove_1ormore_underline(userid)

        self._remove_children(userid)

        self._remove_0ormore_arrow(userid)

        self._remove_1ormore_star(userid)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import requests
from programy.utils.logging.ylogger import YLogger
from programy.services.service import Service
from programy.config.brain.service import BrainServiceConfiguration


class RestAPI:

    def get(self, url, headers=None):
        if headers is None:
            return requests.get(url)
        else:
            return requests.get(url, headers=headers)

    def post(self, url, data, headers=None):
        if headers is None:
            return requests.post(url, data=data)
        else:
            return requests.post(url, data=data, headers=headers)


class GenericRESTService(Service):

    def __init__(self, config: BrainServiceConfiguration, api=None):
        Service.__init__(self, config)

        if api is None:
            self.api = RestAPI()
        else:
            self.api = api

        if config.method is None:
            self.method = "GET"
        else:
            self.method = config.method

        self.host = None
        if config.host is not None:
            self.host = config.host

        self.port = None
        if config.port is not None:
            self.port = config.port

        self.url = None
        if config.url is not None:
            self.url = config.url

    def _format_url(self):
        if self.port is not None:
            host_port = "http://%s:%s" % (self.host, self.port)
        else:
            host_port = "http://%s" % self.host

        if self.url is not None:
            return "%s%s" % (host_port, self.url)
        else:
            return host_port

    def _format_payload(self, client_context, question):
        del client_context
        del question
        return {}

    def _format_get_url(self, url, client_context, question, lang=None, location=None, api_key=None):
        del client_context
        del question
        del lang
        del location
        del api_key
        return url

    def _parse_response(self, client_context, text):
        del client_context
        return text

    def ask_question(self, client_context, question: str):

        try:
            url = self._format_url()

            if self.method == 'GET':
                full_url = self._format_get_url(url, client_context, question)
                response = self.api.get(full_url)
            elif self.method == 'POST':
                payload = self._format_payload(client_context, question)
                response = self.api.post(url, data=payload)
            else:
                raise Exception("Unsupported REST method [%s]" % self.method)

            if response.status_code != 200:
                YLogger.error(client_context, "[%s] return status code [%d]", self.host, response.status_code)
            else:
                return self._parse_response(client_context, response.text)

        except Exception as excep:
            YLogger.exception(client_context, "Failed to resolve", excep)

        return ""



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.utils.logging.ylogger import YLogger

from programy.config.container import BaseContainerConfigurationData
from programy.config.brain.brain import BrainConfiguration
from programy.config.bot.spelling import BotSpellingConfiguration
from programy.config.bot.conversations import BotConversationsConfiguration
from programy.config.bot.splitter import BotSentenceSplitterConfiguration
from programy.config.bot.joiner import BotSentenceJoinerConfiguration
from programy.config.bot.translation import BotTranslatorConfiguration
from programy.config.bot.sentiment import BotSentimentAnalyserConfiguration
from programy.utils.substitutions.substitues import Substitutions


class BotConfiguration(BaseContainerConfigurationData):
    DEFAULT_ROOT = "."
    DEFAULT_RESPONSE = ""
    DEFAULT_RESPONSE_SRAI = ""
    DEFAULT_EMPTY_STRING = ""
    DEFAULT_EXIT_RESPONSE = "Bye!"
    DEFAULT_EXIT_RESPONSE_SRAI = ""
    DEFAULT_INITIAL_QUESTION = "Hello"
    DEFAULT_INITIAL_QUESTION_SRAI = ""
    DEFAULT_OVERRIDE_PREDICATES = True
    DEFAULT_MAX_QUESTION_RECURSION = 100
    DEFAULT_MAX_QUESTION_TIMEOUT = -1
    DEFAULT_MAX_SEARCH_DEPTH = 100
    DEFAULT_MAX_SEARCH_TIMEOUT = -1
    DEFAULT_TAB_PARSE_OUTPUT = True

    def __init__(self, section_name="bot"):

        self._bot_root = BotConfiguration.DEFAULT_ROOT
        self._default_response = BotConfiguration.DEFAULT_RESPONSE
        self._default_response_srai = BotConfiguration.DEFAULT_RESPONSE_SRAI
        self._exit_response = BotConfiguration.DEFAULT_EXIT_RESPONSE
        self._exit_response_srai = BotConfiguration.DEFAULT_EXIT_RESPONSE_SRAI
        self._initial_question = BotConfiguration.DEFAULT_INITIAL_QUESTION
        self._initial_question_srai = BotConfiguration.DEFAULT_INITIAL_QUESTION_SRAI
        self._empty_string = BotConfiguration.DEFAULT_EMPTY_STRING
        self._override_properties = BotConfiguration.DEFAULT_OVERRIDE_PREDICATES
        self._max_question_recursion = BotConfiguration.DEFAULT_MAX_QUESTION_RECURSION
        self._max_question_timeout = BotConfiguration.DEFAULT_MAX_QUESTION_TIMEOUT
        self._max_search_depth = BotConfiguration.DEFAULT_MAX_SEARCH_DEPTH
        self._max_search_timeout = BotConfiguration.DEFAULT_MAX_SEARCH_TIMEOUT
        self._tab_parse_output = BotConfiguration.DEFAULT_TAB_PARSE_OUTPUT
        self._spelling = BotSpellingConfiguration()
        self._from_translator = BotTranslatorConfiguration(name="from_translator")
        self._to_translator = BotTranslatorConfiguration(name="to_translator")
        self._sentiment = BotSentimentAnalyserConfiguration()
        self._conversations = BotConversationsConfiguration()
        self._splitter = BotSentenceSplitterConfiguration()
        self._joiner = BotSentenceJoinerConfiguration()

        brain_config = BrainConfiguration('brain')
        self._brain_configs = [brain_config]
        self._brain_selector = "programy.brainfactory.DefaultBrainSelector"

        BaseContainerConfigurationData.__init__(self, section_name)

    def load_configuration(self, configuration_file, bot_root, subs: Substitutions = None):
        pass

    def load_configuration_section(self, configuration_file, bot, bot_root, subs: Substitutions = None):

        self._default_response = configuration_file.get_option(bot, "default_response",
                                                               BotConfiguration.DEFAULT_RESPONSE, subs=subs)
        self._default_response_srai = configuration_file.get_option(bot, "default_response_srai",
                                                                    BotConfiguration.DEFAULT_RESPONSE_SRAI,
                                                                    subs=subs)
        self._empty_string = configuration_file.get_option(bot, "empty_string",
                                                           BotConfiguration.DEFAULT_EMPTY_STRING, subs=subs)
        self._exit_response = configuration_file.get_option(bot, "exit_response",
                                                            BotConfiguration.DEFAULT_EXIT_RESPONSE, subs=subs)
        self._exit_response_srai = configuration_file.get_option(bot, "exit_response_srai",
                                                                 BotConfiguration.DEFAULT_EXIT_RESPONSE_SRAI,
                                                                 subs=subs)
        self._initial_question = configuration_file.get_option(bot, "initial_question",
                                                               BotConfiguration.DEFAULT_INITIAL_QUESTION, subs=subs)
        self._initial_question_srai = configuration_file.get_option(bot, "initial_question_srai",
                                                                    BotConfiguration.DEFAULT_INITIAL_QUESTION_SRAI,
                                                                    subs=subs)
        self._override_properties = configuration_file.get_option(bot, "override_properties",
                                                                  BotConfiguration.DEFAULT_OVERRIDE_PREDICATES,
                                                                  subs=subs)
        self._max_question_recursion = configuration_file.\
            get_int_option(bot, "max_question_recursion", BotConfiguration.DEFAULT_MAX_QUESTION_RECURSION,
                           subs=subs)
        self._max_question_timeout = configuration_file.get_int_option(bot, "max_question_timeout",
                                                                       BotConfiguration.DEFAULT_MAX_QUESTION_TIMEOUT,
                                                                       subs=subs)
        self._max_search_depth = configuration_file.get_int_option(bot, "max_search_depth",
                                                                   BotConfiguration.DEFAULT_MAX_SEARCH_DEPTH,
                                                                   subs=subs)
        self._max_search_timeout = configuration_file.get_int_option(bot, "max_search_timeout",
                                                                     BotConfiguration.DEFAULT_MAX_SEARCH_TIMEOUT,
                                                                     subs=subs)
        self._tab_parse_output = configuration_file.get_bool_option(bot, "tab_parse_output",
                                                                    BotConfiguration.DEFAULT_TAB_PARSE_OUTPUT,
                                                                    subs=subs)

        self._spelling.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._conversations.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._splitter.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._joiner.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._from_translator.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._to_translator.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self._sentiment.load_config_section(configuration_file, bot, bot_root, subs=subs)

        self.load_brain_configurations(configuration_file, bot, bot_root, subs)

    def load_brain_configurations(self, configuration_file, bot, bot_root, subs: Substitutions = None):
        brains = configuration_file.get_section("brains", bot)
        if brains is not None:
            brain_keys = configuration_file.get_keys(brains)
            first = True
            for name in brain_keys:
                if first is True:
                    self._brain_configs.clear()
                    first = False
                brain_config = configuration_file.get_section(name, brains)
                config = BrainConfiguration(name)
                config.load_configuration(configuration_file, brain_config, bot_root, subs=subs)
                self._brain_configs.append(config)

        self._brain_selector = configuration_file.get_option(bot, "brain_selector", subs=subs)

    @property
    def configurations(self):
        return self._brain_configs

    @property
    def brain_selector(self):
        return self._brain_selector

    @property
    def bot_root(self):
        return self._bot_root

    @property
    def default_response(self):
        return self._default_response

    @default_response.setter
    def default_response(self, text):
        self._default_response = text

    @property
    def default_response_srai(self):
        return self._default_response_srai

    @default_response_srai.setter
    def default_response_srai(self, text):
        self._default_response_srai = text

    @property
    def empty_string(self):
        return self._empty_string

    @empty_string.setter
    def empty_string(self, text):
        self._empty_string = text

    @property
    def exit_response(self):
        return self._exit_response

    @exit_response.setter
    def exit_response(self, text):
        self._exit_response = text

    @property
    def exit_response_srai(self):
        return self._exit_response_srai

    @exit_response_srai.setter
    def exit_response_srai(self, text):
        self._exit_response_srai = text

    @property
    def initial_question(self):
        return self._initial_question

    @initial_question.setter
    def initial_question(self, text):
        self._initial_question = text

    @property
    def initial_question_srai(self):
        return self._initial_question_srai

    @initial_question_srai.setter
    def initial_question_srai(self, text):
        self._initial_question_srai = text

    @property
    def override_properties(self):
        return self._override_properties

    @override_properties.setter
    def override_properties(self, override):
        self._override_properties = override

    @property
    def max_question_recursion(self):
        return self._max_question_recursion

    @property
    def max_question_timeout(self):
        return self._max_question_timeout

    @property
    def max_search_depth(self):
        return self._max_search_depth

    @property
    def max_search_timeout(self):
        return self._max_search_timeout

    @property
    def tab_parse_output(self):
        return self._tab_parse_output

    @property
    def spelling(self):
        return self._spelling

    @property
    def conversations(self):
        return self._conversations

    @property
    def splitter(self):
        return self._splitter

    @property
    def joiner(self):
        return self._joiner

    @property
    def from_translator(self):
        return self._from_translator

    @property
    def to_translator(self):
        return self._to_translator

    @property
    def sentiment_analyser(self):
        return self._sentiment

    def to_yaml(self, data, defaults=True):

        data['bot_root'] = self.bot_root
        data['default_response'] = self.default_response
        data['default_response_srai'] = self.default_response_srai
        data['exit_response'] = self.exit_response
        data['exit_response_srai'] = self.exit_response_srai
        data['initial_question'] = self.initial_question
        data['initial_question_srai'] = self.initial_question_srai
        data['empty_string'] = self.empty_string
        data['override_properties'] = self.override_properties
        data['max_question_recursion'] = self.max_question_recursion
        data['max_question_timeout'] = self.max_question_timeout
        data['max_search_depth'] = self.max_search_depth
        data['max_search_timeout'] = self.max_search_timeout
        data['tab_parse_output'] = self.tab_parse_output
        self.config_to_yaml(data, BotSpellingConfiguration(), defaults)
        self.config_to_yaml(data, BotConversationsConfiguration(), defaults)
        self.config_to_yaml(data, BotSentenceSplitterConfiguration(), defaults)
        self.config_to_yaml(data, BotSentenceJoinerConfiguration(), defaults)
        self.config_to_yaml(data, BotTranslatorConfiguration(name="from_translator"), defaults)
        self.config_to_yaml(data, BotTranslatorConfiguration(name="to_translator"), defaults)
        self.config_to_yaml(data, BotSentimentAnalyserConfiguration(), defaults)

        data['brains'] = {}
        if defaults is True:
            data['brains']['brain'] = {}
            brain_config = BrainConfiguration('brain')
            brain_config.to_yaml(data['brains']['brain'], defaults)

            data['brain_selector'] = "programy.brainfactory.DefaultBrainSelector"

        else:
            for brainconfig in self._brain_configs:
                data['brains'][brainconfig.id] = {}
                brainconfig.to_yaml( data['brains'][brainconfig.id], defaults)

            data['brain_selector'] = self._brain_selector



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.pattern.nodes.base import PatternNode
from programy.parser.pattern.equalsmatch import EqualsMatch
from programy.parser.exceptions import ParserException


class PatternBotNode(PatternNode):

    def __init__(self, attribs, text, userid='*'):
        PatternNode.__init__(self, userid)
        if 'name' in attribs:
            self._property = attribs['name']
        elif 'property' in attribs:
            self._property = attribs['property']
        elif text:
            self._property = text
        else:
            raise ParserException("Invalid bot node, neither name or property specified as attribute or text")

    def is_bot(self):
        return True

    @property
    def property(self):
        return self._property

    def to_xml(self, client_context, include_user=False):
        string = ""
        if include_user is True:
            string += '<bot userid="%s" property="%s">\n' % (self.userid, self.property)
        else:
            string += '<bot property="%s">\n' % self.property
        string += super(PatternBotNode, self).to_xml(client_context)
        string += "</bot>"
        return string

    def to_string(self, verbose=True):
        if verbose is True:
            return "BOT [%s] [%s] property=[%s]" % (self.userid, self._child_count(verbose), self.property)
        return "BOT property=[%s]" % (self.property)

    def equivalent(self, other):
        if other.is_bot():
            if self.userid == other.userid:
                if self.property == other.property:
                    return True
        return False

    def equals(self, client_context, words, word_no):
        word = words.word(word_no)

        if self.userid != '*':
            if self.userid != client_context.userid:
                return EqualsMatch(False, word_no)

        if client_context.brain.properties.has_property(self.property):
            if word == client_context.brain.properties.property(self.property):
                YLogger.debug(client_context, "Found word [%s] as bot property", word)
                return EqualsMatch(True, word_no, word)

        return EqualsMatch(False, word_no)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.utils.text.text import TextUtils
from programy.parser.pattern.match import Match
from programy.parser.pattern.equalsmatch import EqualsMatch
from programy.utils.console.console import outputLog


#######################################################################################################################
#
class MultiValueDict(dict):

    def __setitem__(self, key, value):
        """add the given value to the list of values for this key"""
        self.setdefault(key, []).append(value)

    def remove(self, key, value):
        if key in self:
            if value in self[key]:
                self[key].remove(value)

            if len(self[key]) == 0:
                del self[key]


class PatternNode:
    THAT = "__THAT__"
    TOPIC = "__TOPIC__"

    def __init__(self, userid='*'):

        self._userid = userid

        # Child Nodes
        self._priority_words = []
        self._0ormore_hash = None
        self._1ormore_underline = None
        self._children = []
        self._children_words = MultiValueDict()
        self._iset_names = MultiValueDict()
        self._set_names = MultiValueDict()
        self._bot_properties = MultiValueDict()
        self._0ormore_arrow = None
        self._1ormore_star = None

        # Topic, That and Template Nodes
        self._topic = None
        self._that = None
        self._template = None

    @property
    def userid(self):
        return self._userid

    ########################################################################
    #
    def is_root(self):
        return False

    ########################################################################
    #
    @property
    def priority_words(self):
        return self._priority_words

    def has_priority_words(self):
        return True if self._priority_words else False

    def is_priority(self):
        return False

    ########################################################################
    #
    @property
    def arrow(self):
        return self._0ormore_arrow

    @property
    def hash(self):
        return self._0ormore_hash

    @property
    def underline(self):
        return self._1ormore_underline

    @property
    def star(self):
        return self._1ormore_star

    def has_wildcard(self):
        if self.has_zero_or_more() is True or self.has_one_or_more():
            return True
        return False

    def has_zero_or_more(self):
        return bool(self._0ormore_arrow is not None or self._0ormore_hash is not None)

    def has_0ormore_arrow(self):
        return bool(self._0ormore_arrow is not None)

    def has_0ormore_hash(self):
        return bool(self._0ormore_hash is not None)

    def has_one_or_more(self):
        return bool(self._1ormore_star is not None or self._1ormore_underline is not None)

    def has_1ormore_star(self):
        return bool(self._1ormore_star is not None)

    def has_1ormore_underline(self):
        return bool(self._1ormore_underline is not None)

    def is_wildcard(self):
        return False

    def is_zero_or_more(self):
        return False

    def is_one_or_more(self):
        return False

    def is_word(self):
        return False

    ########################################################################
    #
    @property
    def children(self):
        return self._children

    def child(self, num):
        return self._children[num]

    def has_children(self):
        if self._children:
            return True
        return False

    def has_nodes(self):
        if self._children:
            return True

        if self.has_priority_words():
            return True

        if self.has_wildcard():
            return True

        return False

    ########################################################################
    #
    @property
    def topic(self):
        return self._topic

    def is_topic(self):
        return False

    def has_topic(self):
        return bool(self.topic is not None)

    def add_topic(self, topic):
        if self.has_topic() is False:
            self._topic = topic
        return self._topic

    ########################################################################
    #
    @property
    def that(self):
        return self._that

    def add_that(self, that):
        if self.has_that() is False:
            self._that = that
        return self._that

    def is_that(self):
        return False

    def has_that(self):
        return bool(self.that is not None)

    ########################################################################
    #
    @property
    def template(self):
        return self._template

    def is_template(self):
        return False

    def has_template(self):
        return bool(self.template is not None)

    def add_template(self, template):
        if self.has_template() is False:
            self._template = template
        return self._template

    ########################################################################
    #
    def is_set(self):
        return False

    def is_iset(self):
        return False

    ########################################################################
    #
    def is_bot(self):
        return False

    ########################################################################
    #
    def is_regex(self):
        return False

    ########################################################################
    #
    def equivalent(self, other):
        del other
        return False

    def equals(self, client_context, words, word_no):
        del words
        del client_context
        return EqualsMatch(False, word_no)

    def equals_ignore_case(self, word1, word2):
        if word1 is not None and word2 is not None:
            return bool(word1.upper() == word2.upper())
        return False

    ########################################################################
    #
    def can_add(self, new_node):
        pass    # pragma: no cover

    def _priority_node_exist(self, new_node):
        for priority in self._priority_words:
            if priority.equivalent(new_node) is True:
                # Equivalent node already exists, use this one instead
                return priority
        return None

    def _zero_or_more_node_exist(self, new_node):
        if self._0ormore_arrow is not None:
            if self._0ormore_arrow.equivalent(new_node) is True:
                return self._0ormore_arrow
        if self._0ormore_hash is not None:
            if self._0ormore_hash.equivalent(new_node) is True:
                return self._0ormore_hash
        return None

    def _one_or_more_node_exist(self, new_node):
        if self._1ormore_underline is not None:
            if self._1ormore_underline.equivalent(new_node) is True:
                return self._1ormore_underline
        if self._1ormore_star is not None:
            if self._1ormore_star.equivalent(new_node) is True:
                return self._1ormore_star
        return None

    def _topic_node_exist(self, new_node):
        del new_node
        if self._topic is not None:
            return self._topic
        return None

    def _that_node_exist(self, new_node):
        del new_node
        if self._that is not None:
            return self._that
        return None

    def _template_node_exist(self, new_node):
        del new_node
        if self._template is not None:
            return self._template
        return None

    def _iset_node_exist(self, new_node):
        for existing_node in self.children:
            if existing_node.is_iset():
                if existing_node.equivalent(new_node):
                    return existing_node
        return None

    def _set_node_exist(self, new_node):
        if new_node.set_name in self._set_names:
            existing_nodes = self._set_names[new_node.set_name]
            for existing_node in existing_nodes:
                if existing_node.equivalent(new_node):
                    # Equivalent node already exists, use this one instead
                    return existing_node
        return None

    def _bot_node_exist(self, new_node):
        if new_node.property in self._bot_properties:
            existing_nodes = self._bot_properties[new_node.property]
            for existing_node in existing_nodes:
                if existing_node.equivalent(new_node):
                    # Equivalent node already exists, use this one instead
                    return existing_node
        return None

    def _regex_node_exist(self, new_node):
        for existing_node in self.children:
            if existing_node.is_regex():
                if existing_node.equivalent(new_node):
                    return existing_node
        return None

    def _word_node_exist(self, new_node):
        if new_node.word in self._children_words:
            existing_nodes = self._children_words[new_node.word]
            for existing_node in existing_nodes:
                if existing_node.equivalent(new_node):
                    # Equivalent node already exists, use this one instead
                    return existing_node
        return None

    def _node_exists(self, new_node):

        if new_node.is_priority() is True:
            return self._priority_node_exist(new_node)

        if new_node.is_zero_or_more() is True:
            return self._zero_or_more_node_exist(new_node)

        if new_node.is_one_or_more() is True:
            return self._one_or_more_node_exist(new_node)

        if new_node.is_topic() is True:
            return self._topic_node_exist(new_node)

        if new_node.is_that():
            return self._that_node_exist(new_node)

        if new_node.is_template() is True:
            return self._template_node_exist(new_node)

        if new_node.is_iset() is True:
            return self._iset_node_exist(new_node)

        if new_node.is_set() is True:
            return self._set_node_exist(new_node)

        if new_node.is_bot() is True:
            return self._bot_node_exist(new_node)

        if new_node.is_regex() is True:
            return self._regex_node_exist(new_node)

        if new_node.is_word():
            return self._word_node_exist(new_node)

        return None

    def _add_node(self, new_node):

        # Otherwise use the new node, and return that to maintain consistence
        # And allow child node to be chained, but supports duplicates
        if new_node.is_priority() is True:
            self._priority_words.append(new_node)

        elif new_node.is_zero_or_more() is True:
            if new_node.wildcard == '^':
                self._0ormore_arrow = new_node
            else:
                self._0ormore_hash = new_node

        elif new_node.is_one_or_more() is True:
            if new_node.wildcard == '_':
                self._1ormore_underline = new_node
            else:
                self._1ormore_star = new_node

        elif new_node.is_template() is True:
            self._template = new_node

        else:
            # Append sets and bots to the end of the array as they take a slightly
            # lower priority to actual words.
            # This allows the following to work
            #  my favorite color is green
            #  my favorite color is <set>color</set>
            # In the above, if the set color contains green then
            # it still gets picked up in the first grammar and not he second
            if new_node.is_set() is True:
                self.children.append(new_node)
                self._set_names[new_node.set_name] = new_node

            elif new_node.is_iset() is True:
                self.children.append(new_node)
                self._iset_names[new_node.iset_name] = new_node

            elif new_node.is_bot() is True:
                self.children.append(new_node)
                self._bot_properties[new_node.property] = new_node

            elif new_node.is_regex() is True:
                self.children.append(new_node)

            else:
                self.children.insert(0, new_node)
                if new_node.is_word() is True:
                    self._children_words[new_node.word] = new_node

        return new_node

    def _remove_node(self, current_node):
        YLogger.debug(None, "Removing %s" % current_node.to_string())

        if current_node.is_priority() is True:
            self._priority_words.remove(current_node)

        elif current_node.is_zero_or_more() is True:
            if current_node.wildcard == '^':
                self._0ormore_arrow = None
            else:
                self._0ormore_hash = None

        elif current_node.is_one_or_more() is True:
            if current_node.wildcard == '_':
                self._1ormore_underline = None
            else:
                self._1ormore_star = None

        elif current_node.is_template() is True:
            self._template = None

        else:
            # Append sets and bots to the end of the array as they take a slightly
            # lower priority to actual words.
            # This allows the following to work
            #  my favorite color is green
            #  my favorite color is <set>color</set>
            # In the above, if the set color contains green then
            # it still gets picked up in the first grammar and not he second
            if current_node.is_set() is True:
                self.children.remove(current_node)
                self._set_names.remove(current_node.set_name, current_node)

            elif current_node.is_iset() is True:
                self.children.remove(current_node)
                self._iset_names.remove(current_node.iset_name, current_node)

            elif current_node.is_bot() is True:
                self.children.remove(current_node)
                self._bot_properties.remove(current_node.property, current_node)

            elif current_node.is_regex() is True:
                self.children.remove(current_node)

            else:
                # Assumption is that node is Word Node, if none of the above
                self.children.remove(current_node)
                self._children_words.remove(current_node.word, current_node)

    def add_child(self, new_node, replace_existing=False):

        # Check the rules allow this now to be a child of the current node
        self.can_add(new_node)

        # Next check for duplicates, returning original if one exists
        exists = self._node_exists(new_node)
        if exists is not None and replace_existing is False:
            return exists

        # Otherwise add the new node to the appropriate container
        result = self._add_node(new_node)
        return result

    def _child_count(self, verbose=True):
        if verbose is True:
            return "P(%d)^(%d)#(%d)C(%d)_(%d)*(%d)To(%d)Th(%d)Te(%d)" % (
                len(self._priority_words),
                1 if self._0ormore_arrow is not None else 0,
                1 if self._0ormore_hash is not None else 0,
                len(self.children),
                1 if self._1ormore_underline is not None else 0,
                1 if self._1ormore_star is not None else 0,
                1 if self._topic is not None else 0,
                1 if self._that is not None else 0,
                1 if self._template is not None else 0
            )
        return ""

    def to_string(self, verbose=True):
        if verbose is True:
            return "NODE [%s] [%s]" % (self.userid, self._child_count(verbose))
        return "NODE"

    def get_tabs(self, client_context, depth):
        if client_context.bot.configuration.tab_parse_output is True:
            return TextUtils.get_tabs(depth)
        return ""

    def dump(self, tabs, output_func=YLogger.debug, eol="", verbose=True):

        string = "{0}{1}{2}".format(tabs, self.to_string(verbose), eol)
        if output_func == outputLog or YLogger.is_ylogger_method(output_func) is True:
            output_func(self, string)
        else:
            output_func(string)

        for priority in self._priority_words:
            priority.dump(tabs + "\t", output_func, eol, verbose)

        if self._0ormore_arrow is not None:
            self._0ormore_arrow.dump(tabs + "\t", output_func, eol, verbose)

        if self._0ormore_hash is not None:
            self._0ormore_hash.dump(tabs + "\t", output_func, eol, verbose)

        if self._1ormore_underline is not None:
            self._1ormore_underline.dump(tabs + "\t", output_func, eol, verbose)

        if self._1ormore_star is not None:
            self._1ormore_star.dump(tabs + "\t", output_func, eol, verbose)

        if self._topic is not None:
            self._topic.dump(tabs + "\t", output_func, eol, verbose)

        if self._that is not None:
            self._that.dump(tabs + "\t", output_func, eol, verbose)

        if self._template is not None:
            self._template.dump(tabs + "\t", output_func, eol, verbose)

        for child in self.children:
            child.dump(tabs + "\t", output_func, eol, verbose)

    def to_xml(self, client_context, include_user=False):
        string = ""

        for priority in self._priority_words:
            string += priority.to_xml(client_context, include_user)

        if self._0ormore_arrow is not None:
            string += self._0ormore_arrow.to_xml(client_context, include_user)

        if self._0ormore_hash is not None:
            string += self._0ormore_hash.to_xml(client_context, include_user)

        if self._1ormore_underline is not None:
            string += self._1ormore_underline.to_xml(client_context, include_user)

        if self._1ormore_star is not None:
            string += self._1ormore_star.to_xml(client_context, include_user)

        if self._topic is not None:
            string += self._topic.to_xml(client_context, include_user)

        if self._that is not None:
            string += self._that.to_xml(client_context, include_user)

        if self._template is not None:
            string += self._template.to_xml(client_context)

        for child in self.children:
            string += child.to_xml(client_context, include_user)

        return string

    def match_children(self, client_context, children, child_type, words, word_no, context, match_type, depth):

        tabs = self.get_tabs(client_context, depth)

        for child in children:

            result = child.equals(client_context, words, word_no)
            if result.matched is True:
                word_no = result.word_no
                YLogger.debug(client_context, "%s%s matched %s", tabs, child_type, result.matched_phrase)

                match_node = Match(match_type, child, result.matched_phrase)
                context.add_match(match_node)

                match = child.consume(client_context, context, words, word_no + 1, match_type, depth + 1)
                if match is not None:
                    YLogger.debug(client_context, "%sMatched %s child, success!", tabs, child_type)
                    return match, word_no

                else:
                    context.pop_match()

        return None, word_no

    def consume(self, client_context, context, words, word_no, match_type, depth, parent=False):
        del parent

        tabs = self.get_tabs(client_context, depth)

        if context.search_time_exceeded() is True:
            YLogger.error(client_context, "%sMax search time [%d]secs exceeded", tabs, context.max_search_timeout)
            return None

        if context.search_depth_exceeded(depth) is True:
            YLogger.error(client_context, "%sMax search depth [%d] exceeded", tabs, context.max_search_depth)
            return None

        if word_no >= words.num_words():
            if self._template is not None:
                YLogger.debug(client_context, "%sFound a template, success!", tabs)
                return self._template
            else:
                YLogger.debug(client_context, "%sNo more words and no template, no match found!", tabs)
                return None

        if self._topic is not None:
            match = self._topic.consume(client_context, context, words, word_no, Match.TOPIC, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched topic, success!", tabs)
                return match

            if words.word(word_no) == PatternNode.TOPIC:
                YLogger.debug(client_context, "%s Looking for a %s, none give, no match found!", tabs,
                              PatternNode.TOPIC)
                return None

        if self._that is not None:
            match = self._that.consume(client_context, context, words, word_no, Match.THAT, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched that, success!", tabs)
                return match

            if words.word(word_no) == PatternNode.THAT:
                YLogger.debug(client_context, "%s Looking for a %s, none give, no match found!", tabs, PatternNode.THAT)
                return None

        match, word_no = self.match_children(client_context, self._priority_words, "Priority", words, word_no, context,
                                             match_type, depth)
        if match is not None:
            return match

        if self._0ormore_hash is not None:
            match = self._0ormore_hash.consume(client_context, context, words, word_no, match_type, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched 0 or more hash, success!", tabs)
                return match

        if self._1ormore_underline is not None:
            match = self._1ormore_underline.consume(client_context, context, words, word_no, match_type, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched 1 or more underline, success!", tabs)
                return match

        match, word_no = self.match_children(client_context, self._children, "Word", words, word_no, context,
                                             match_type, depth)
        if match is not None:
            return match

        if self._0ormore_arrow is not None:
            match = self._0ormore_arrow.consume(client_context, context, words, word_no, match_type, depth + 1)
            YLogger.debug(client_context, "%sMatched 0 or more arrow, success!", tabs)
            return match

        if self._1ormore_star is not None:
            match = self._1ormore_star.consume(client_context, context, words, word_no, match_type, depth + 1)
            if match is not None:
                YLogger.debug(client_context, "%sMatched 1 or more star, success!", tabs)
                return match

        YLogger.debug(client_context, "%sNo match for %s, trying another path", tabs, words.word(word_no))
        return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.file.store.filestore import FileStore
from programy.storage.entities.category import CategoryReadOnlyStore


class FileCategoryStore(FileStore, CategoryReadOnlyStore):

    def __init__(self, storage_engine):
        FileStore.__init__(self, storage_engine)
        CategoryReadOnlyStore.__init__(self)

    def _get_storage_path(self):
        return self.storage_engine.configuration.categories_storage.dirs

    def get_storage(self):
        return self.storage_engine.configuration.categories_storage

    def load_all(self, collector):
        dirs = self.storage_engine.configuration.categories_storage.dirs
        cat_ext = self.storage_engine.configuration.categories_storage.extension
        subdirs = self.storage_engine.configuration.categories_storage.subdirs

        if self.storage_engine.configuration.categories_storage.has_single_file():
            for filename in dirs:
                if cat_ext is not None:
                    if filename.endswith(cat_ext):
                        self._load_file_contents(collector, filename)

                else:
                    self._load_file_contents(collector, filename)
        else:
            for cat_dir in dirs:
                if subdirs is False:
                    paths = os.listdir(cat_dir)
                    for filename in paths:
                        if filename.endswith(cat_ext):
                            self._load_file_contents(collector, os.path.join(cat_dir, filename))

                else:
                    for dirpath, _, filenames in os.walk(cat_dir):
                        for filename in [f for f in filenames if f.endswith(cat_ext)]:
                            self._load_file_contents(collector, os.path.join(dirpath, filename))

    def load(self, collector, filename):
        self._load_file_contents(collector, filename)

    def _load_file_contents(self, collection, filename):
        YLogger.debug(self, "Loading file contents from [%s]", filename)
        try:
            collection.parse_from_file(filename, userid="*")

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to load cat [%s]", excep, filename)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.section import BaseSectionConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class BrainDynamicsConfiguration(BaseSectionConfigurationData):

    def __init__(self):
        BaseSectionConfigurationData.__init__(self, "dynamic")
        self._dynamic_sets = {}
        self._dynamic_maps = {}
        self._dynamic_vars = {}

    @property
    def dynamic_sets(self):
        return self._dynamic_sets

    @property
    def dynamic_maps(self):
        return self._dynamic_maps

    @property
    def dynamic_vars(self):
        return self._dynamic_vars

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        dynamic_config = configuration_file.get_section("dynamic", configuration)
        if dynamic_config is not None:
            self.load_dynamic_sets(configuration_file, dynamic_config, subs=subs)
            self.load_dynamic_maps(configuration_file, dynamic_config, subs=subs)
            self.load_dynamic_vars(configuration_file, dynamic_config, subs=subs)
        else:
            YLogger.error(self, "Config section [dynamic] missing from Brain, using defaults")

    def load_dynamic_sets(self, configuration_file, dynamic_config, subs: Substitutions = None):
        sets_config = configuration_file.get_option(dynamic_config, "sets", subs=subs)
        if sets_config is not None:
            for set_key in sets_config.keys():
                dyn_set_class = sets_config[set_key]
                self._dynamic_sets[set_key.upper()] = dyn_set_class

    def load_dynamic_maps(self, configuration_file, dynamic_config, subs: Substitutions = None):
        maps_config = configuration_file.get_option(dynamic_config, "maps", subs=subs)
        if maps_config is not None:
            for map_name in maps_config.keys():
                dyn_map_class = maps_config[map_name]
                self._dynamic_maps[map_name.upper()] = dyn_map_class

    def load_dynamic_vars(self, configuration_file, dynamic_config, subs: Substitutions = None):
        vars_config = configuration_file.get_option(dynamic_config, "variables", subs=subs)
        if vars_config is not None:
            for var_name in vars_config.keys():
                dyn_var_class = vars_config[var_name]
                self._dynamic_vars[var_name.upper()] = dyn_var_class

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['sets'] = {}
            data['sets']['NUMBER'] = 'programy.dynamic.sets.numeric.IsNumeric'
            data['sets']['ROMAN'] = 'programy.dynamic.sets.roman.IsRomanNumeral'
            data['sets']['STOPWORD'] = 'programy.dynamic.sets.stopword.IsStopWord'
            data['sets']['SYNSETS'] = 'programy.dynamic.sets.synsets.IsSynset'

            data['maps'] = {}
            data['maps']['ROMANTODDEC'] = 'programy.dynamic.maps.roman.MapRomanToDecimal'
            data['maps']['DECTOROMAN'] = 'programy.dynamic.maps.roman.MapDecimalToRoman'
            data['maps']['LEMMATIZE'] = 'programy.dynamic.maps.lemmatize.LemmatizeMap'
            data['maps']['STEMMER'] = 'programy.dynamic.maps.stemmer.StemmerMap'

            data['variables'] = {}
            data['variables']['GETTIME'] = 'programy.dynamic.variables.datetime.GetTime'

        else:
            data['sets'] = {}
            for key, value in self._dynamic_sets.items():
                data['sets'][key] = value

            data['maps'] = {}
            for key, value in self._dynamic_maps.items():
                data['maps'][key] = value

            data['variables'] = {}
            for key, value in self._dynamic_vars.items():
                data['variables'][key] = value
